import urllib.request

# the 'certifi' library is required at APL and other sites that
# require SSL certs for web fetches.  If you need this, install certifi
# (pip install certifi)
import importlib

certspec = importlib.util.find_spec("certifi")
found = certspec is not None
if found: import certifi

import pandas as pd  # dataframes and also to_datetime
import json
import re
import datetime

"""
;supermag_api.py
; ================
; Author S. Antunes, based on supermag-api.pro by R.J.Barnes


; (c) 2021  The Johns Hopkins University Applied Physics Laboratory
;LLC.  All Rights Reserved. 

;This material may be only be used, modified, or reproduced by or for
;the U.S. Government pursuant to the license rights granted under the 
;clauses at DFARS 252.227-7013/7014 or FAR 52.227-14. For any other
;permission, 
;please contact the Office of Technology Transfer at JHU/APL.

; NO WARRANTY, NO LIABILITY. THIS MATERIAL IS PROVIDED "AS IS."
; JHU/APL MAKES NO REPRESENTATION OR WARRANTY WITH RESPECT TO THE
; PERFORMANCE OF THE MATERIALS, INCLUDING THEIR SAFETY, EFFECTIVENESS, 
; OR COMMERCIAL VIABILITY, AND DISCLAIMS ALL WARRANTIES IN THE
; MATERIAL, WHETHER EXPRESS OR IMPLIED, INCLUDING (BUT NOT LIMITED TO)
; ANY AND ALL IMPLIED WARRANTIES OF PERFORMANCE, MERCHANTABILITY,
; FITNESS FOR A PARTICULAR PURPOSE, AND NON-INFRINGEMENT OF
; INTELLECTUAL PROPERTY OR OTHER THIRD PARTY RIGHTS. ANY USER OF THE
; MATERIAL ASSUMES THE ENTIRERISK AND LIABILITY FOR USING THE
; MATERIAL. IN NO EVENT SHALL JHU/APL BE LIABLE TO ANY USER OF THE
; MATERIAL FOR ANY ACTUAL, INDIRECT, CONSEQUENTIAL, SPECIAL OR OTHER
; DAMAGES ARISING FROM THE USE OF, OR INABILITY TO USE, THE MATERIAL,
; INCLUDING, BUT NOT LIMITED TO, ANY DAMAGES FOR LOST PROFITS. 
"""


# Sample URLs, type into browser if you want to compare the data vs python
# https://supermag.jhuapl.edu/services/data-api.php?fmt=json&logon=YOURNAME&start=2019-10-15T10:40&extent=3600&all&station=NCK
# https://supermag.jhuapl.edu/services/indices.php?fmt=json&logon=YOURNAME&start=2019-10-15T10:40&extent=3600&all
# https://supermag.jhuapl.edu/services/inventory.php?fmt=json&logon=YOURNAME&start=2019-10-15T10:40&extent=3600


def sm_coreurl(page, logon, start, extent):
    # internal helper function
    baseurl = "https://supermag.jhuapl.edu/"

    mytime = sm_parsestart(start)
    urlstr = baseurl + 'services/' + page + '?python&nohead'
    urlstr += '&start=' + mytime
    urlstr += '&logon=' + logon
    urlstr += '&extent=' + ("%12.12d" % extent)

    # print("debug:",urlstr)

    return (urlstr)


# handy helper function when using complicated CVS encoding of lists
def sm_csvitem_to_list(myarr):
    # converts entity of form ['HOP', 'NVS', 'IRT'] to an actual list of HOP, NVS, IRT
    mylist = []
    for myline in myarr:
        myline = re.sub("'", "", myline[1:-1])
        mylist.append(myline.split(", "))
    return (mylist)


# handy helper function when using complicated CVS encoding of dicts
def sm_csvitem_to_dict(myarr, **kwargs):
    # converts entity of form {'X': -12.213, 'Y': -5.5, 'Z': 1.2} to an actual dict of var.X, var.Y, etc
    # items are presumed strings by default
    # optional argument "convert=num" will convert them to doubles
    mylist = []
    for myline in myarr:
        myline2 = re.sub(" ", "", myline[1:-1])  # scrub out extra spaces
        myline2 = re.sub("'", "", myline2)
        elements = dict(x.split(":") for x in myline2.split(","))
        # little sanity check to make sure float subitems remain floats
        try:
            elements = {item: float(value) for (item, value) in elements.items()}
        except:
            pass

        # type(elements)
        mylist.append(elements)
    return (mylist)


def sm_parsestart(start):
    # internal helper function
    # takes either list of [yyyy, mo, dd, hh, mm, opt_ss]
    # or string of a normal datetime 'YYYY-MM-DD hh-mm' (optional ss)
    # or the SuperMAG-ready 'YYYY-MM-DDThh-mm-ss'

    if isinstance(start, list):
        timestring = "%4.4d-%2.2d-%2.2dT%2.2d:%2.2d" % tuple(start[0:5])
    elif isinstance(start, datetime.date):
        # good to go, TBD
        timestring = start.strftime("%Y-%m-%dT%H:%M")
    else:
        # is a string, reparse, TBD
        timestring = start

    return (timestring)


def sm_DateToYMDHMS(tval, yr, mo, dy, hr, mt, sc):
    # not used but here as an example of date conversion
    julday = (tval / 86400.0) + 2440587.5
    datestr = pd.to_datetime(julday, unit='D', origin='julian')  # format YYYY-MM-DD HH:MM:SS.ssssss
    return (datestr)


def sm_keycheck_data(flagstring):
    # internal helper function
    toggles = ['mlt', 'mag', 'geo', 'decl', 'sza', 'delta=start', 'baseline=yearly', 'baseline=none']

    myflags = ''
    flags = [x.strip() for x in flagstring.split(',')]

    for i in range(0, len(flags)):
        chk = flags[i]
        chk = chk.lower()
        # check for the '*all', also individual keys, and assemble url flags
        if chk == 'all': myflags += '&mlt&mag&geo&decl&sza'
        for ikey in range(0, len(toggles)):
            if chk == toggles[ikey]: myflags += '&' + toggles[ikey]

    return (myflags)


def sm_keycheck_indices(flagstring):
    # internal helper function
    # For category='indices', always returns:
    #        tval
    # additional flags to return data include:
    #        indicesall (or its alias: all)
    #  (or any of)
    #        baseall, sunall, darkall, regionalall, plusall
    #  (or specify individual items to include, from the sets below)
    #
    basekeys = ["sme", "sml", "smu", "mlat", "mlt", "glat", "glon", "stid", "num"]
    # sunkeys: alias allowed of SUN___ -> ___s
    sunkeys = ["smes", "smls", "smus", "mlats", "mlts", "glats", "glons", "stids", "nums"]
    # darkkeys: alias allowed of DARK___ -> ___d
    darkkeys = ["smed", "smld", "smud", "mlatd", "mltd", "glatd", "glond", "stidd", "numd"]
    # regkeys: alias allowed of REGIONAL___ -> ___r
    regkeys = ["smer", "smlr", "smur", "mlatr", "mltr", "glatr", "glonr", "stidr", "numr"]
    pluskeys = ["smr", "ltsmr", "ltnum", "nsmr"]
    indiceskeys = basekeys + sunkeys + darkkeys + regkeys + pluskeys
    # 'all' means all the above

    imfkeys = ["bgse", "bgsm", "vgse", "vgsm"]  # or imfall for all these
    swikeys = ["pdyn", "epsilon", "newell", "clockgse", "clockgsm",
               "density"]  # % or swiall for all these
    myflags = ''
    indices = '&indices='
    swi = '&swi='
    imf = '&imf='

    flags = [x.strip() for x in flagstring.split(',')]

    for i in range(0, len(flags)):
        chk = flags[i]
        chk = chk.lower()

        # check for the '*all', also individual keys, and assemble url flags
        if chk == 'all': indices += 'all,'
        if chk == 'indicesall': indices += 'all,'
        if chk == 'imfall': imf += 'all,'
        if chk == 'swiall': swi += 'all,'
        # available keywords, we allow both the url version and the
        # aliases of "SUN___ -> ___s", "DARK___ -> ___d", "REGIONAL___ -> ___r"

        for ikey in range(0, len(indiceskeys)):
            mykey = indiceskeys[ikey]
            sunkey = "sun" + mykey  # allow alias
            darkkey = "dark" + mykey  # allow alias
            regkey1 = "regional" + mykey  # allow alias
            regkey2 = "reg" + mykey  # allow alias
            if chk == mykey:
                indices += mykey + ','  # base key is correct
            elif sunkey == mykey:
                indices += mykey + 's,'  # alias, so base key + 's'
            elif darkkey == mykey:
                indices += mykey + 'd,'  # alias, so base key + 'd'
            elif regkey1 == mykey or regkey2 == mykey:
                indices += mykey + 'r,'  # alias, so base key + 'r'

        for ikey in range(0, len(swikeys)):
            if chk == swikeys[ikey]: swi += swikeys[ikey] + ','

        for ikey in range(0, len(imfkeys)):
            if chk == imfkeys[ikey]: imf += imfkeys[ikey] + ','

        # more aliases to the user
        if chk == 'baseall': indices += ','.join(basekeys)
        if chk == 'sunall': indices += ','.join(sunkeys)
        if chk == 'darkall': indices += ','.join(darkkeys)
        if chk == 'regionalall' or chk == 'regall': indices += ','.join(regkeys)
        if chk == 'plusall': indices += ','.join(pluskeys)

    # clean it up a bit by removing extraneous tags/characters
    if indices == "&indices=": indices = ""
    if swi == "&swi=": swi = ""
    if imf == "&imf=": imf = ""
    # add them together
    myflags = indices + swi + imf
    # a little more cleaning for tidiness, removes extraneous commas
    myflags = re.sub(',&', '&', myflags)
    myflags = re.sub(',$', '', myflags)

    return (myflags)


def sm_GetUrl(fetchurl, fetch='raw'):
    # internal helper function
    # returned data choices are 'raw' or 'json', default is 'raw'
    # converts an http bytestream into a python list (raw) or list of dict (json)
    # 'stations' should be 'raw' and returns a list
    # 'data' should be 'json', returns a list (which converts to a dataframe)
    # 'indices' should be 'json', returns a list (which converts to a dataframe)

    success = 0  # gets changed to 1 good data is fetched
    longstring = b"ERROR: Unknown error"  # prepare for the worst

    # print("debug: url trying ",fetch,"is",fetchurl)
    # If the url object throws an error it will be caught here
    try:
        # with urllib.request.urlopen(fetchurl,cafile=certifi.where()) as response:
        # note that 'cafile' is deprecated past python 3.5 but we keep it here
        # to have stronger backward compatability with earlier versions
        try:
            cafile = certifi.where()
        except:
            cafile = ''
        # with urllib.request.urlopen(fetchurl, cafile=cafile) as response:
        with urllib.request.urlopen(fetchurl,) as response:
            longstring = response.read()

            if fetch == 'json':
                if len(longstring) > 3:
                    # mydata = json.loads(longstring[3:]) # skipping initial OK
                    mydata = json.loads(longstring)
                else:
                    mydata = []  # just the word 'OK', no data, so return no data
                success = 1
            else:
                # default is raw byte strings, which we split into a list
                mydata = (longstring.decode('UTF-8')).split('\n')
                success = 1  # it worked
                if re.search(r'ERROR', mydata[0]): success = 0  # legit return, but of an err

    except urllib.error.URLError as e:
        # print, !ERROR_STATE.msg
        mydata = ['ERROR:HTTP error', e.reason]
    except:
        longstring = longstring.decode('UTF-8')
        mydata = [longstring]  # catch-all if nothing below works

    # print("debug: function return type is:",type(mydata),".")
    # returns a list for 'raw' or a list of dictionaries for 'json'
    return (success, mydata)


# Gets a list of stations.
# Return value is either '1' plus list of stations, or
#                        '0' plus a string with the error message
# Sample usage:
# 'extent' is how long a window, in seconds, to grab. (86400 sec = 1 day)
# (status, stations) =supermaggetinventory('myname',2019,11,2,20,24,00,86400)
# In this case, 'status'=1 and 'stations' is a list of 184 stations

def SuperMAGGetInventory(logon, start, extent):
    # One of the core 3 functions

    iarr = ""
    errstr = ""

    # construct URL
    urlstr = sm_coreurl('inventory.php', logon, start, extent)

    # get the string array of stations
    (success, stations) = sm_GetUrl(urlstr, 'raw')

    # if the inventory is valid extract the stations to an array
    # if an error occurs set the the ERROR keyword to be the error string

    if success:
        # first data item is how many stations were found
        numstations = int(stations[0])
        # print("debug: found",numstations,"stations")
        if numstations > 0:
            stations = stations[1:-1]  # remove OK,#stations,''
        else:
            stations = []  # empty list because no stations found
        # print('debug, got back:',stations)
    # success: return true (1) if the call was successful otherwise false (0)
    # stations: return list with true/1 plus data, or false/0 plus errstr
    return (success, stations)


def SuperMAGGetIndices(logon, start, extent, flagstring='', **kwargs):
    # One of the core 3 functions

    urlstr = sm_coreurl('indices.php', logon, start, extent)
    indices = sm_keycheck_indices(flagstring)
    urlstr += indices

    # get the string array of JSON data
    (status, data_list) = sm_GetUrl(urlstr, 'json')

    # default is to return a dataframe, but can also return an array
    if (kwargs.get('FORMAT', 'none')).lower() == 'list':
        return (status, data_list)
    else:
        # default, converts the json 'list of dictionaries' into a dataframe
        data_df = pd.DataFrame(data_list)
        return (status, data_df)


def SuperMAGGetData(logon, start, extent, flagstring, station, **kwargs):
    # One of the core 3 functions

    # optional options for 'data':
    # ALL=&mlt&mag&geo&decl&sza
    # MLT=&mlt,MAG=&mag,GEO=&geo,DECL=&decl,SZA=&sza,
    # DELTA='start',BASELINE='none/yearly'
    # e.g. can pass  MLT=1,MAG=1  and they will be evaluated.  Full set checked: ALL, MLT, MAG, GEO, DECL, SZA, also values for DELTA, BASELINE
    # also arg FORMAT='list', otherwise defaults to FORMAT='dataframe'  NOT YET DONE!!!!

    # default FORMAT='dataframe', alt is FORMAT='list'

    urlstr = sm_coreurl('data-api.php', logon, start, extent)
    indices = sm_keycheck_data(flagstring)
    urlstr += indices
    urlstr += '&station=' + station.upper()

    (status, data_list) = sm_GetUrl(urlstr, 'json')

    # default is to return a dataframe, but can also return an array
    if (kwargs.get('FORMAT', 'none')).lower() == 'list':
        return (status, data_list)
    else:
        # default, converts the json 'list of dictionaries' into a dataframe
        data_df = pd.DataFrame(data_list)
        return (status, data_df)


def sm_grabme(dataf, key, subkey):
    # syntactical sugar to grab nested subitems from a dataframe
    data = dataf[key]
    subdata = [temp[subkey] for temp in data]
    return (subdata)


# Unlike IDL, which returns as Array or Struct,
# we return as List (of dictionaries) or DataFrame

def sm_microtest(choice, userid):
    # 3 simple unit tests to verify the core fetches work
    import matplotlib.pyplot as plt

    start = [2019, 11, 15, 10, 40, 00]  # alt: start='2019-11-15T10:40'

    if choice == 1 or choice == 4:
        (status, stations) = SuperMAGGetInventory(userid, start, 3600)
        print(status)
        print(stations)

    if choice == 2 or choice == 4:
        (status, data) = SuperMAGGetData(userid, start, 3600, 'all,delta=start,baseline=yearly', 'HBK')
        print(status)
        print(data)
        print(data.keys())

        tval = data.tval
        mlt = data.mlt
        ### Python way
        N_nez = [temp['nez'] for temp in data.N]
        N_geo = [temp['geo'] for temp in data.N]
        ### or, supermag helper shorthand way
        N_nez = sm_grabme(data, 'N', 'nez')
        N_geo = sm_grabme(data, 'N', 'geo')
        #
        plt.plot(tval, N_nez)
        plt.plot(tval, N_geo)
        plt.ylabel('N_geo vs N_nez')
        plt.xlabel('date')
        plt.show()

    if choice == 3 or choice == 4:
        (status, idxdata) = SuperMAGGetIndices(userid, start, 3600, 'swiall,density,darkall,regall,smes')
        # print(status)
        # print(idxdata)
        idxdata.keys()
        tval = idxdata.tval
        hours = list(range(24))
        y = idxdata.SMLr
        for i in range(len(tval) - 1):
            plt.plot(hours, y[i])
            plt.ylabel('SMLr')
            plt.xlabel('hour')
            plt.title('SMLr variation by hour, for successive days')
        plt.show()


def supermag_testing(userid):
    start = [2019, 11, 15, 10, 40, 00]  # alt: start='2019-11-15T10:40'

    (status, stations) = SuperMAGGetInventory(userid, start, 3600)

    # DATA fetches
    # BARE CALL, dataframe returned
    (status, mydata1a) = SuperMAGGetData(userid, start, 3600, '', 'HBK')
    mydata1a  # is 1440 rows x 6 columns dataframe
    mydata1a.keys()  # Index(['tval', 'ext', 'iaga', 'N', 'E', 'Z'], dtype='object')

    # CALL with ALLINDICES, dataframe returned
    (status, mydata1a) = SuperMAGGetData(userid, start, 3600, 'all', 'HBK')
    mydata1a  # is 1440 rows x 12 columns dataframe
    mydata1a.keys()  # Index(['tval', 'ext', 'iaga', 'glon', 'glat', 'mlt', 'mcolat', 'decl', 'sza', 'N', 'E', 'Z'], dtype='object')

    # BARE CALL, list returned
    (status, mydata1b) = SuperMAGGetData(userid, start, 3600, '', 'HBK', FORMAT='list')
    len(mydata1b)  # is 1440 rows of dicts (key-value pairs)
    mydata1b[
    0:1]  # {'tval': 1572726240.0, 'ext': 60.0, 'iaga': 'DOB', 'N': {'nez': -3.942651, 'geo': -5.964826}, 'E': {'nez': 4.492887, 'geo': 0.389075}, 'Z': {'nez': 7.608168, 'geo': 7.608168}}

    # CALL with ALLINDICES, list returned
    (status, mydata1b) = SuperMAGGetData(userid, start, 3600, 'all', 'HBK', FORMAT='list')
    mydata1b  # is 1440 rows of dicts (key-value pairs)
    mydata1b[
    0:1]  # {'tval': 1572726240.0, 'ext': 60.0, 'iaga': 'DOB', 'glon': 9.11, 'glat': 62.07, 'mlt': 21.694675, 'mcolat': 30.361519, 'decl': 3.067929, 'sza': 124.698227, 'N': {'nez': -3.942651, 'geo': -5.964826}, 'E': {'nez': 4.492887, 'geo': 0.389075}, 'Z': {'nez': 7.608168, 'geo': 7.608168}}

    ####################
    # INDICES fetches
    (status, idxdata) = SuperMAGGetIndices(userid, start, 3600)
    idxdata  # empty!

    (status, idxdata) = SuperMAGGetIndices(userid, start, 3600, 'all,swiall,imfall')
    idxdata  # 1440 rows x 77 columns dataframe
    idxdata.keys()  # Index(['tval', 'SME', 'SML', 'SMLmlat', 'SMLmlt', 'SMLglat', 'SMLglon', 'SMLstid', 'SMU', 'SMUmlat', 'SMUmlt', 'SMUglat', 'SMUglon', 'SMUstid', 'SMEnum', 'SMEs', 'SMLs', 'SMLsmlat', 'SMLsmlt', 'SMLsglat', 'SMLsglon', 'SMLsstid', 'SMUs', 'SMUsmlat', 'SMUsmlt', 'SMUsglat', 'SMUsglon', 'SMUsstid', 'SMEsnum', 'SMEd', 'SMLd', 'SMLdmlat', 'SMLdmlt', 'SMLdglat', 'SMLdglon', 'SMLdstid', 'SMUd', 'SMUdmlat', 'SMUdmlt', 'SMUdglat', 'SMUdglon', 'SMUdstid', 'SMEdnum', 'SMEr', 'SMLr', 'SMLrmlat', 'SMLrmlt', 'SMLrglat', 'SMLrglon', 'SMLrstid', 'SMUr', 'SMUrmlat', 'SMUrmlt', 'SMUrglat', 'SMUrglon', 'SMUrstid', 'SMErnum', 'smr', 'smr00', 'smr06', 'smr12', 'smr18', 'smrnum', 'smrnum00', 'smrnum06', 'smrnum12', 'smrnum18', 'bgse', 'bgsm', 'vgse', 'vgsm', 'clockgse', 'clockgsm', 'density', 'dynpres', 'epsilon', 'newell'], dtype='object')
    #
    # just INDICESALL = 67 columns, above 'tval' through 'smrnum18'
    # just IMFALL = 5 columns, Index(['tval', 'bgse', 'bgsm', 'vgse', 'vgsm'], dtype='object')
    # just SWIALL = 7 columns, Index(['tval', 'clockgse', 'clockgsm', 'density', 'dynpres', 'epsilon', 'newell'], dtype='object')
    #
    # Dataframes are awesome!  To manipulate, just pull out what you need
    import pandas as pd  # call once at the top of your code if you are using dataframes
    tval = idxdata.tval
    density = idxdata.density
    vgse = idxdata.vgse
    # or all as 1 line of code
    tval, density, vgse = idxdata.tval, idxdata.density, idxdata.vgse
    # note that vgse is itself a dictionary of values for X/Y/Z, so you can get subitems from it like this
    vgse_x = [d.get('X') for d in idxdata.vgse]

    # to save the data, there are many formats.  Here is how to save as csv
    idxdata.to_csv('mydata.csv')

    # to read it back in later
    import pandas as pd
    import re
    mydata2b = pd.read_csv('mydata.csv',
                           index_col=0)  # you can read it into any variable name, we just used 'mydata2b' as an example
    # now you can do all the above items again, with one exception: each line of the CVS file got split into a dict (key-value pairs) but items like 'vsge' are part of the pandas structure
    # the 'd.get()' approach will _not_ work once read from csv
    stationlist = mydata2b.SMLrstid  # item is a pandas series (not python list)
    print(stationlist[
              0])  # prints a list of stations as a string, but cannot easily access a single item because it is a pandas series
    # so you can convert a pandas series to a list
    stationlist2 = sm_csvitem_to_list(mydata2b.SMLrstid)  # goal is a list of stations
    slist = stationlist2[0]  # grabs a list of stations for row 0
    s1 = stationlist2[0][0]  # grabs the first station for row 0

    vgse = sm_csvitem_to_dict(mydata2b.vgse)  # goal is a dict of coords or other values
    x = vgse[0]['X']  # grab just the 'X' value for the 1st row of data
    vgse_x = [mydat['X'] for mydat in vgse]  # grab all the 'X' values as a new list
    vgse_xyz = [(mydat['X'], mydat['Y'], mydat['Z']) for mydat in vgse]  # grab all 3

    # We also offer a list format, for users who prefer to work in python lists
    (status, mydata2c) = SuperMAGGetIndices(userid, start, 3600, 'all,swiall,imfall', FORMAT='list')
    len(mydata2c)  # is 1440 rows of dicts (key-value pairs)
    mydata2c[
    0:1]  # {'tval': 1572726240.0, 'SME': 58.887299, 'SML': -27.709055, 'SMLmlat': 73.529922, 'SMLmlt': 23.321493, 'SMLglat': 76.510002, 'SMLglon': 25.01, 'SMLstid': 'HOP', 'SMU': 31.178246, 'SMUmlat': 74.702339, 'SMUmlt': 2.090216, 'SMUglat': 79.480003, 'SMUglon': 76.980003, 'SMUstid': 'VIZ', 'SMEnum': 118, 'SMEs': 34.451469, 'SMLs': -16.599854, 'SMLsmlat': 62.368008, 'SMLsmlt': 9.399416, 'SMLsglat': 62.299999, 'SMLsglon': 209.800003, 'SMLsstid': 'T39', 'SMUs': 17.851616, 'SMUsmlat': 73.989975, 'SMUsmlt': 18.228394, 'SMUsglat': 67.93, 'SMUsglon': 306.429993, 'SMUsstid': 'ATU', 'SMEsnum': 54, 'SMEd': 58.887299, 'SMLd': -27.709055, 'SMLdmlat': 73.529922, 'SMLdmlt': 23.321493, 'SMLdglat': 76.510002, 'SMLdglon': 25.01, 'SMLdstid': 'HOP', 'SMUd': 31.178246, 'SMUdmlat': 74.702339, 'SMUdmlt': 2.090216, 'SMUdglat': 79.480003, 'SMUdglon': 76.980003, 'SMUdstid': 'VIZ', 'SMEdnum': 64, 'SMEr': [29.685059, 29.857538, 31.387127, 41.707573, 10.320444, 10.885443, 9.604616, 13.479583, 15.471248, 15.471248, 15.714731, 5.434914, 12.13654, 11.156847, 9.62884, 14.752592, 14.752592, 24.204388, 21.41181, 21.41181, 27.121195, 46.345322, 51.403328, 51.403328], 'SMLr': [-27.709055, 1.320708, -0.208882, -10.529325, -10.529325, -10.529325, -9.248499, -13.123466, -16.599854, -16.599854, -16.599854, -5.449972, -5.449972, -4.470279, -2.942272, -6.352773, -6.352773, -6.352773, -3.560194, -3.560194, -7.514064, -22.651047, -27.709055, -27.709055], 'SMLrmlat': [73.529922, 51.264774, 47.791527, 66.696564, 66.696564, 66.696564, 41.771515, 70.602707, 62.368008, 62.368008, 62.368008, 67.471809, 67.471809, 60.639145, 68.500282, 72.20977, 72.20977, 72.20977, 75.762718, 75.762718, 77.33667, 71.889503, 73.529922, 73.529922], 'SMLrmlt': [23.321493, 2.119074, 3.578985, 4.929673, 4.929673, 4.929673, 5.414416, 8.57761, 9.399416, 9.399416, 9.399416, 11.35623, 11.35623, 12.266475, 13.977451, 16.720993, 16.720993, 16.720993, 19.65963, 19.65963, 21.307804, 22.863134, 23.321493, 23.321493], 'SMLrglat': [76.510002, 55.029999, 52.169998, 71.580002, 71.580002, 71.580002, 47.799999, 71.300003, 62.299999, 62.299999, 62.299999, 61.756001, 61.756001, 53.351002, 58.763, 63.75, 63.75, 63.75, 72.300003, 72.300003, 76.769997, 74.5, 76.510002, 76.510002], 'SMLrglon': [25.01, 82.900002, 104.449997, 129.0, 129.0, 129.0, 132.414001, 203.25, 209.800003, 209.800003, 209.800003, 238.770004, 238.770004, 247.026001, 265.920013, 291.480011, 291.480011, 291.480011, 321.700012, 321.700012, 341.369995, 19.200001, 25.01, 25.01], 'SMLrstid': ['HOP', 'NVS', 'IRT', 'TIK', 'TIK', 'TIK', 'BRN', 'BRW', 'T39', 'T39', 'T39', 'FSP', 'FSP', 'C06', 'FCC', 'IQA', 'IQA', 'IQA', 'SUM', 'SUM', 'DMH', 'BJN', 'HOP', 'HOP'], 'SMUr': [1.976003, 31.178246, 31.178246, 31.178246, -0.208882, 0.356117, 0.356117, 0.356117, -1.128606, -1.128606, -0.885122, -0.015059, 6.686568, 6.686568, 6.686568, 8.399819, 8.399819, 17.851616, 17.851616, 17.851616, 19.60713, 23.694275, 23.694275, 23.694275], 'SMUrmlat': [52.904049, 74.702339, 74.702339, 74.702339, 47.791527, 54.29908, 54.29908, 54.29908, 66.244217, 66.244217, 57.76614, 54.597057, 55.715378, 55.715378, 55.715378, 57.829525, 57.829525, 73.989975, 73.989975, 73.989975, 70.473801, 68.194489, 68.194489, 68.194489], 'SMUrmlt': [0.510692, 2.090216, 2.090216, 2.090216, 3.578985, 6.394085, 6.394085, 6.394085, 9.99274, 9.99274, 11.729218, 12.269058, 13.969843, 13.969843, 13.969843, 16.160952, 16.160952, 18.228394, 18.228394, 18.228394, 21.200783, 22.967857, 22.967857, 22.967857], 'SMUrglat': [56.432999, 79.480003, 79.480003, 79.480003, 52.169998, 59.970001, 59.970001, 59.970001, 64.047997, 64.047997, 51.882999, 47.664001, 45.870998, 45.870998, 45.870998, 48.650002, 48.650002, 67.93, 67.93, 67.93, 70.900002, 71.089996, 71.089996, 71.089996], 'SMUrglon': [58.567001, 76.980003, 76.980003, 76.980003, 104.449997, 150.860001, 150.860001, 150.860001, 220.889999, 220.889999, 239.973999, 245.791, 264.916992, 264.916992, 264.916992, 287.549988, 287.549988, 306.429993, 306.429993, 306.429993, 351.299988, 25.790001, 25.790001, 25.790001], 'SMUrstid': ['ARS', 'VIZ', 'VIZ', 'VIZ', 'IRT', 'MGD', 'MGD', 'MGD', 'DAW', 'DAW', 'C13', 'C10', 'C08', 'C08', 'C08', 'T50', 'T50', 'ATU', 'ATU', 'ATU', 'JAN', 'NOR', 'NOR', 'NOR'], 'SMErnum': [5, 3, 3, 4, 5, 6, 6, 4, 8, 9, 12, 13, 20, 17, 17, 11, 12, 14, 12, 14, 22, 51, 51, 35], 'smr': 0.252399, 'smr00': -0.531382, 'smr06': 0.885406, 'smr12': 1.051192, 'smr18': -0.395618, 'smrnum': 72, 'smrnum00': 26, 'smrnum06': 23, 'smrnum12': 6, 'smrnum18': 17, 'bgse': {'X': 1.07, 'Y': -3.75, 'Z': -0.74}, 'bgsm': {'X': 1.07, 'Y': -3.82, 'Z': -0.06}, 'vgse': {'X': -351.100006, 'Y': -5.5, 'Z': -4.0}, 'vgsm': {'X': 351.100006, 'Y': 6.128625, 'Z': -2.947879}, 'clockgse': 258.340698, 'clockgsm': 268.664337, 'density': 5.03, 'dynpres': 1.25, 'epsilon': 29.468521, 'newell': 2504.155029}
    # sample accessing
    print(mydata2c[0]['tval'], mydata2c[0]['density'])  # single element
    result = [(myeach['tval'], myeach['density']) for myeach in mydata2c]  # pull out pairs e.g. 'tval, density')
    # two-line method for extracting any variable set from this
    pairsets = [(myeach['tval'], myeach['density'], myeach['vgse']) for myeach in
                mydata2c]  # same, pull out pairs, only assign e.g. x=tval, y=density
    tval, density, vgse = [[z[i] for z in pairsets] for i in (0, 1, 2)]
    # since 'vgse' is itself an dict of 3 values X/Y/Z, you can pull out nested items like this
    pairsets = [(myeach['tval'], myeach['density'], myeach['vgse']['X']) for myeach in
                mydata2c]  # same, pull out pairs, only assign e.g. x=tval, y=density
    tval, density, vgse_x = [[z[i] for z in pairsets] for i in (0, 1, 2)]
    # the above methods are extensible to any number of variables, just update the (0,1,2) to reflect now many you have

#  Uncomment to run quick sample tests
# userid=YOUR_SUPERMAG_USER_ID
# sm_microtest(1,userid)   # sample stations fetch
# sm_microtest(2,userid)   # sample data fetch, with plotting
# sm_microtest(3,userid)   # sample indices fetch, with plotting


def test_indices():
    userid = 'JouleCai'
    start = [2019, 11, 15, 10, 40, 00]
    (status, idxdata) = SuperMAGGetIndices(userid, start, 3600, 'all,swiall,imfall')
    idxdata  # 1440 rows x 77 columns dataframe
    idxdata.keys()

if __name__ == "__main__":

    test_indices()


# Licensed under the BSD 3-Clause License
# Copyright (C) 2021 GeospaceLab (geospacelab)
# Author: Lei Cai, Space Physics and Astronomy, University of Oulu

__author__ = "Lei Cai"
__copyright__ = "Copyright 2021, GeospaceLab"
__license__ = "BSD-3-Clause License"
__email__ = "lei.cai@oulu.fi"
__docformat__ = "reStructureText"


import datetime
import numpy as np
import requests
import bs4
import pathlib
import re
import netCDF4 as nc
import ftplib
from contextlib import closing

import geospacelab.toolbox.utilities.pydatetime as dttool
import geospacelab.toolbox.utilities.pylogging as mylog
import geospacelab.datahub.sources.wdc as wdc
from geospacelab.config import prf


class Downloader(object):

    def __init__(
            self, dt_fr,  dt_to,
            data_file_root_dir=None, force=False, ftp_sub_dir=None, ftp_filename_prefix=None,
            ftp_filename_ext='txt'
    ):

        self.dt_fr = dt_fr
        self.dt_to = dt_to
        self.force = force
        self.done = False
        if data_file_root_dir is None:
            self.data_file_root_dir = prf.datahub_data_root_dir / 'GFZ' / 'Indices'
        else:
            self.data_file_root_dir = pathlib.Path(data_file_root_dir)

        self.ftp_host = "ftp.gfz-potsdam.de"
        self.ftp_subdir = ftp_sub_dir
        self.ftp_filename_prefix = ftp_filename_prefix
        self.ftp_filename_ext = ftp_filename_ext

    def download(self):
        diff_years = self.dt_to.year - self.dt_fr.year

        for i in range(diff_years + 1):
            dt1 = datetime.datetime(self.dt_fr.year + i, 1, 1)

            ystr = dt1.strftime('%Y')

            with closing(ftplib.FTP()) as ftp:
                try:
                    ftp.connect(self.ftp_host, 21, 30)  # 30 timeout
                    ftp.login()
                    ftp.cwd('/pub/home/obs/' + self.ftp_subdir)

                    file_list = ftp.nlst()

                    file_name_patterns = list([self.ftp_filename_prefix+ystr])

                    file_name = self.search_files(file_list=file_list, file_name_patterns=file_name_patterns)

                    file_path = self.data_file_root_dir / file_name

                    if file_path.is_file() and not self.force:
                        mylog.simpleinfo.info(
                            "The file {} exists in the directory {}.".format(file_path.name, file_path.parent.resolve()))
                        self.done = True
                    else:
                        file_path.parent.resolve().mkdir(parents=True, exist_ok=True)
                        with open(file_path, 'w+b') as f:
                            res = ftp.retrbinary('RETR ' + file_name, f.write)

                            if not res.startswith('226 Transfer complete'):
                                print('Downloaded of file {0} is not compile.'.format(file_name))
                                pathlib.Path.unlink(file_path)
                                self.done = False
                                return None
                except:
                    print('Error during download from FTP')
                    self.done = False
                    return

            mylog.StreamLogger.info("Preparing to save the data in the netcdf format ...")
            self.save_to_netcdf(ystr, file_path)

    @staticmethod
    def search_files(file_list=None, file_name_patterns=None):

        if file_name_patterns is None:
            file_name_patterns = []

        # filter with file name patterns
        if list(file_name_patterns):
            search_pattern = '.*' + '.*'.join(file_name_patterns) + '.*'
            fn_regex = re.compile(search_pattern)
            file_list = list(filter(fn_regex.match, file_list))

        if not list(file_list):
            return None

        if len(file_list) > 1:
            raise ValueError

        return file_list[0]

    def save_to_netcdf(self, *args, **kwargs):
        pass


if __name__ == "__main__":
    dt_fr1 = datetime.datetime(2020, 1, 1)
    dt_to1 = datetime.datetime(2020, 12, 16)
    Downloader(dt_fr1, dt_to1)






# Licensed under the BSD 3-Clause License
# Copyright (C) 2021 GeospaceLab (geospacelab)
# Author: Lei Cai, Space Physics and Astronomy, University of Oulu

__author__ = "Lei Cai"
__copyright__ = "Copyright 2021, GeospaceLab"
__license__ = "BSD-3-Clause License"
__email__ = "lei.cai@oulu.fi"
__docformat__ = "reStructureText"


import sys
import inspect

import geospacelab.toolbox.utilities.pylogging as mylog
import geospacelab.toolbox.utilities.pyclass as pyclass
import weakref
import numpy as np
from geopack import geopack

default_coord_attrs = {
    'sph':  {
        'lat':  {'unit': 'deg'},
        'lon':  {'unit': 'deg'},
        'phi':  {'unit': 'rad'},
        'theta': {'unit': 'rad'},
        'height': {'unit': 'km'},
        'r':    {'unit': 'Re'},
        'mlt':  {'unit': 'h'},
        'lst':  {'unit': 'h'},
        'az':   {'unit': 'deg'},
        'el':   {'unit': 'deg'},
        'range': {'unit':'deg'},
    },
    'car':  {
        'x':    {'unit': 'Re'},
        'y':    {'unit': 'Re'},
        'z':    {'unit': 'Re'},
        'E':    {'unit': 'km'},
        'N':    {'unit': 'km'},
        'Up':   {'unit': 'km'},
        'Down': {'unit': 'km'},
        'Forward':  {'unit': 'km'},
        'Perp':     {'unit': 'km'},
    }
}


class SpaceCSBase(object):
    def __init__(self,  name=None, vector=None, coords=None, ut=None, kind=None, new_coords=None, **kwargs):
        self.name = name
        self.ut = ut
        self.coords = None
        self.vector = vector
        self.kind = kind
        self._set_coords(coords, new_coords=new_coords)

    def __call__(self, *args, cs_to=None, **kwargs):
        if not list(args):
            self.vector = None
        elif len(args) == 1:
            from geospacelab.datahub import VariableModel
            if issubclass(args.__class__, VariableModel):
                self.vector = args[0].value
            else:
                self.vector = args[0]
        elif len(args) == 3:
            self.vector = np.concatenate(args, axis=len(args[0].shape)+1)
        else:
            raise NotImplementedError

        transform_func = getattr(self, 'to_' + cs_to.upper())
        cs_new = transform_func(**kwargs)
        return cs_new

    def _set_coords(self, coords, new_coords=None):
        if self.kind == 'sph':
            self.coords = SphericalCoordinates(cs=self, new_coords=new_coords)
        elif self.kind == 'car':
            self.coords = CartesianCoordinates(cs=self, new_coords=new_coords)

        attrs = {}

        if isinstance(coords, dict):
            attrs = coords
        elif issubclass(coords.__class__, self.coords.__class__):
            self.coords = coords
        for key, value in attrs.items():
            if hasattr(self.coords, key):
                if 'unit' not in key:
                    value = validate_coord_value(value)
                setattr(self.coords, key, value)
            else:
                mylog.StreamLogger.warning(
                    'The coords have not the attr {}. Add a new coord using the method "add_coord"'.format(key)
                )

    def config(self, logging=True, **kwargs):
        pyclass.set_object_attributes(self, append=False, logging=logging, **kwargs)

    def __getitem__(self, key):
        return getattr(self.coords, key)

    def __setitem__(self, key, value):
        if hasattr(self.coords, key):
            setattr(self.coords, key, value)
        else:
            raise KeyError

    def to_GEO(self, cs_new=None, **kwargs):
        if self.name == 'GEO':
            return self
        if cs_new is None:
            raise NotImplementedError
        else:
            return cs_new

    def to_GEOC(self, cs_new=None, **kwargs):
        if self.name == 'GEOC':
            return self

        if cs_new is None:
            raise NotImplementedError
        else:
            return cs_new

    def to_AACGM(self, cs_new=None, **kwargs):
        if self.name == 'AACGM':
            return self
        if cs_new is None:
            raise NotImplementedError
        else:
            return cs_new

    def to_APEX(self, cs_new=None, **kwargs):
        if self.name == 'APEX':
            return self
        if cs_new is None:
            raise NotImplementedError
        else:
            return cs_new

    def to_GEI(self, cs_new=None, **kwargs):
        if self.name == 'GEI':
            return self
        if cs_new is None:
            raise NotImplementedError
        else:
            return cs_new

    def to_SM(self, cs_new=None, **kwargs):
        if self.name == 'SM':
            return self
        if cs_new is None:
            raise NotImplementedError
        else:
            return cs_new

    def to_GSW(self, cs_new=None, **kwargs):
        if self.name == 'GSW':
            return self
        if cs_new is None:
            raise NotImplementedError
        else:
            return cs_new

    def to_QD(self, cs_new=None, **kwargs):
        if self.name == 'QD':
            return self
        if cs_new is None:
            raise NotImplementedError
        else:
            return cs_new


class SpaceSphericalCS(SpaceCSBase):
    def __init__(self,  name=None, coords=None, vector=None, ut=None, new_coords=None, **kwargs):

        super().__init__(name=name, vector=vector, coords=coords, ut=ut, kind='sph', new_coords=new_coords, **kwargs)

    def to_cartesian(self, **kwargs):
        from geospacelab.cs import set_cs

        if self.name in ['AACGM', 'APEX']:
            mylog.StreamLogger.error("{} has the spherical coordinates only!".format(self.name))

        if self.name in ['GEO', 'GEOD']:
            mylog.StreamLogger.warning("{} is converted into geocentric coordinates!".format(self.name))
            cs_new = self.to_GEOC()
            return cs_new.sph_to_car()

        x = self.coords.r * np.sin(self.coords.theta) * np.cos(self.coords.phi)
        y = self.coords.r * np.sin(self.coords.theta) * np.sin(self.coords.phi)
        z = self.coords.r * np.cos(self.coords.theta)

        coords = {'x': x, 'y': y, 'z': z}

        vector = None
        if self.vector is not None:
            mylog.StreamLogger.warning("The tranformation for vectors have not been implemented!")

        cs_new = set_cs(name=self.name, coords=coords, vector=vector, kind='car', ut=self.ut, **kwargs)

        return cs_new


class SpaceCartesianCS(SpaceCSBase):
    def __init__(self,  name=None, coords=None, vector=None, ut=None, new_coords=None, **kwargs):

        super().__init__(name=name, vector=vector, coords=coords, ut=ut, kind='car', new_coords=new_coords, **kwargs)

    def to_spherical(self, **kwargs):
        from geospacelab.cs import set_cs
        import geospacelab.toolbox.utilities.numpymath as npmath

        # phi: longitude, theta: co-latitude, r: radial distance
        r = np.sqrt(self.coords.x ** 2 + self.coords.y ** 2 + self.coords.z ** 2)
        r_unit = self.coords.x_unit
        theta = np.arccos(self.coords.z / r)
        phi = npmath.trig_arctan_to_sph_lon(self.coords.x, self.coords.y)
        vector = None
        if self.vector is not None:
            mylog.StreamLogger.warning("The tranformation for vectors have not been implemented!")

        coords = {'theta': theta, 'phi': phi, 'r': r, 'r_unit': r_unit}
        cs_new = set_cs(name=self.name, coords=coords,  vector=vector, kind='sph', ut=self.ut, **kwargs)

        return cs_new


class CoordinatesBase(object):
    def __init__(self, cs=None, kind=None, new_coords=None):
        self.cs = weakref.proxy(cs)
        self.kind = kind
        if isinstance(new_coords, list):
            for coord_name in new_coords:
                self.add_coord(coord_name)

    def __call__(self, **kwargs):
        self.config(**kwargs)

    def config(self, logging=True, **kwargs):
        pyclass.set_object_attributes(self, append=False, logging=logging, **kwargs)

    def add_coord(self, name, value=None, unit=None):
        if not hasattr(self, name):
            value = validate_coord_value(value)
            setattr(self, name, value)
            if unit is None:
                unit = default_coord_attrs[self.kind][name]['unit']
            setattr(self, name + '_unit', unit)
        else:
            raise AttributeError("The coord is existing!")


class SphericalCoordinates(CoordinatesBase):
    def __init__(self, cs=None, **kwargs):
        self.Re = np.double(6371.2)
        kwargs.setdefault('new_coords', ['r', 'theta', 'phi'])
        super().__init__(cs=cs, kind='sph', **kwargs)


class CartesianCoordinates(CoordinatesBase):
    def __init__(self, cs=None, **kwargs):
        self.Re = np.double(6371.2)     # Earth rad in km
        kwargs.setdefault('new_coords', ['x', 'y', 'z'])
        super().__init__(cs=cs, kind='car', **kwargs)


def validate_coord_value(value):

    if np.issubdtype(type(value), np.number):
        value = value
    elif type(value) in (list, tuple):
        value = np.array(value)
    elif isinstance(value, np.ndarray):
        value = value
    elif value is not None:
        raise TypeError
    return value

# def check_cs_method(cls):
#     sph_methods = ["convert_latlon_to_thetaphi", "convert_thetaphi_to_latlon",
#                    "convert_height_to_r", "convert_r_to_height"]
#     car_methods = ["convert_xyz_to_E_N_Up"]
#
#     def list_func():
#         func_dict = {
#             name: obj
#             for name, obj in inspect.getmembers(sys.modules[__name__])
#             if (inspect.isfunction(obj)) and (obj.__module__ == __name__)
#         }
#         return func_dict
#
#     def bind(instance, method):
#         def binding_scope_fn(*args, **kwargs):
#             return method(instance, *args, **kwargs)
#
#         return binding_scope_fn
#
#     def wrapper(*args, **kwargs):
#         kind = kwargs.setdefault('kind', None)
#         name = kwargs.setdefault('name', None)
#         func_dict = list_func()
#         instance = cls(*args, **kwargs)
#         if kind == "sph":
#             for sph_method in sph_methods:
#                 if name in ["GEO", "GEOD"]:
#                     continue
#                 setattr(instance, sph_method, bind(instance, func_dict[sph_method]))
#         if kind == "car":
#             for car_method in car_methods:
#
#                 setattr(instance, car_method, bind(instance, func_dict[car_method]))
#         return instance
#     return wrapper


# @check_cs_method
# class SpaceCSBase(object):
#     def __init__(self, name=None, coords=None, ut=None, kind=None, new_coords=None, **kwargs):
#
#         self.name = name
#         self.ut = ut
#         self.coords = None
#         self.kind = kind
#         self._set_coords(new_coords=new_coords)
#
#         attrs = {}
#
#         if isinstance(coords, dict):
#             attrs = coords
#         elif issubclass(coords.__class__, self.coords.__class__):
#             self.coords = coords
#
#         for key, value in attrs.items():
#             if hasattr(self.coords, key):
#                 setattr(self.coords, key, value)
#             else:
#                 mylog.StreamLogger.warning(
#                     'The coords have not the attr {}. Add a new coord using the method "add_coord"'.format(key)
#                 )
#
#     def _set_coords(self, new_coords=None):
#         if self.kind == 'sph':
#             self.coords = SphericalCoordinates(cs=self, new_coords=new_coords)
#         elif self.kind == 'car':
#             self.coords = CartesianCoordinates(cs=self, new_coords=new_coords)
#
#     def transform(self, cs_to=None, **kwargs):
#         func = getattr(self, 'to_' + cs_to.upper())
#         cs_new = func(**kwargs)
#         return cs_new
#
#     def sph_to_car(self):
#         if self.kind == 'sph':
#             return self
#
#         if self.name in ['AACGM', 'APEX']:
#             mylog.StreamLogger.error("{} has the spherical coordinates only!".format(self.name))
#
#         if self.name in ['GEO', 'GEOD']:
#             mylog.StreamLogger.warning("{} will be converted into GEOC in advance!".format(self.name))
#             cs_new = self.to_GEOC
#             return cs_new.sph_to_car()
#
#         from geospacelab.cs import set_cs
#         cs_new = set_cs(name=self.name, kind='car', ut=self.ut, new_coords=['x', 'y', 'z'])
#         cs_new['x'] = self.r * np.sin(self.theta) * np.cos(self.phi)
#         cs_new['y'] = self.r * np.sin(self.theta) * np.sin(self.phi)
#         cs_new['z'] = self.r * np.cos(self.theta)
#
#         return cs_new
#
#     def car_to_sph(self):
#         if self.kind == 'car':
#             return self
#         if self.name in ['AACGM', 'APEX', 'GEO', 'GEOD']:
#             mylog.StreamLogger.error("{} has no Cartersian coordinates!".format(self.name))
#
#         import geospacelab.toolbox.utilities.numpymath as npmath
#         from geospacelab.cs import set_cs
#         # phi: longitude, theta: co-latitude, r: radial distance
#         r = np.sqrt(self.x ** 2 + self.y ** 2 + self.z ** 2)
#         theta = np.arccos(self.z / r)
#         phi = npmath.trig_arctan_to_sph_lon(self.x, self.y)
#
#         cs_new = set_cs(name=self.name, kind='sph', ut=self.ut, new_coords=['r', 'theta', 'phi'])
#
#         cs_new['r'] = r
#         cs_new['phi'] = phi
#         cs_new['theta'] = theta
#
#         return cs_new
#
#     def config(self, logging=True, **kwargs):
#         pyclass.set_object_attributes(self, append=False, logging=logging, **kwargs)
#
#     def __getitem__(self, key):
#         return getattr(self.coords, key)
#
#     def __setitem__(self, key, value):
#         if hasattr(self.coords, key):
#             setattr(self.coords, key, value)
#         else:
#             raise KeyError
#
#
# class CoordinatesBase(object):
#     def __init__(self, cs=None, kind=None, new_coords=None):
#         self.cs = weakref.proxy(cs)
#         self.kind = kind
#         if isinstance(new_coords, list):
#             for coord_name in new_coords:
#                 self.add_coord(coord_name)
#
#     def __call__(self, **kwargs):
#         self.config(**kwargs)
#
#     def config(self, logging=True, **kwargs):
#         pyclass.set_object_attributes(self, append=False, logging=logging, **kwargs)
#
#     def add_coord(self, name, value=None, unit=None):
#         if not hasattr(self, name):
#             setattr(self, name, value)
#             if unit is None:
#                 unit = default_coord_attrs[self.kind][name]['unit']
#             setattr(self, name + '_unit', unit)
#         else:
#             raise AttributeError("The coord is existing!")
#
#
# class SphericalCoordinates(CoordinatesBase):
#     def __init__(self, cs=None, **kwargs):
#         self._Re = 6371.2
#         kwargs.setdefault('new_coords', ['r', 'theta', 'phi'])
#         super().__init__(cs=cs, kind='sph', **kwargs)
#
#     def _convert_height_to_r(self):
#         r = (self.height + self._Re) / self._Re
#         self.add_coord('r', value=r)
#         return r
#
#     def _convert_r_to_height(self):
#         if self.r_unit == 'Re':
#             factor = self._Re
#         elif self.r_unit == 'km':
#             factor = 1.
#         else:
#             raise NotImplementedError
#         height = self.r * factor - self._Re
#         self.add_coord('height', value=height)
#         return height
#
#     def _convert_latlon_to_thetaphi(self):
#         if self.lon_unit == 'rad':
#             factor = 1.
#         elif self.phi_unit == 'deg':
#             factor = np.pi / 180.
#         else:
#             raise NotImplementedError
#         theta = 90. - self.lat * factor
#         phi = self.phi * factor
#         self.add_coord('phi', value=phi)
#         self.add_coord('theta', value=theta)
#
#     def
#
#
# def convert_thetaphi_to_latlon(self):
#
#     if self.phi_unit == 'rad':
#         factor = 180. / np.pi
#     elif self.phi_unit == 'deg':
#         factor = 1.
#     else:
#         raise NotImplementedError
#     lat = 90. - self.theta * factor
#     lon = self.phi * factor
#     self.add_coord('lat', value=lat)
#     self.add_coord('lon', value=lon)
#
#
# def convert_height_to_r(self):
#     r = (self.height + self._Re) / self._Re
#     self.add_coord('r', value=r)
#     return r
#
#
# def convert_r_to_height(self):
#     if self.r_unit == 'Re':
#         factor = self._Re
#     elif self.r_unit == 'km':
#         factor = 1.
#     else:
#         raise NotImplementedError
#     height = self.r * factor - self._Re
#     self.add_coord('height', value=height)
#     return height
#
#
# def convert_latlon_to_thetaphi(self):
#     if self.lon_unit == 'rad':
#         factor = 1.
#     elif self.phi_unit == 'deg':
#         factor = np.pi / 180.
#     else:
#         raise NotImplementedError
#     theta = 90. - self.lat * factor
#     phi = self.phi * factor
#     self.add_coord('phi', value=phi)
#     self.add_coord('theta', value=theta)
#
#
# class CartesianCoordinates(CoordinatesBase):
#     def __init__(self, cs=None, **kwargs):
#         self._Re = 6371.2       # Earth rad in km
#         kwargs.setdefault('new_coords', ['x', 'y', 'z'])
#         super().__init__(cs=cs, kind='car', **kwargs)
#
#



# Licensed under the BSD 3-Clause License
# Copyright (C) 2021 GeospaceLab (geospacelab)
# Author: Lei Cai, Space Physics and Astronomy, University of Oulu

__author__ = "Lei Cai"
__copyright__ = "Copyright 2021, GeospaceLab"
__license__ = "BSD-3-Clause License"
__email__ = "lei.cai@oulu.fi"
__docformat__ = "reStructureText"


import datetime
import pathlib

import ftplib
import requests
import bs4
import re
import tqdm

import geospacelab.toolbox.utilities.pydatetime as dttool
import geospacelab.toolbox.utilities.pylogging as mylog
from geospacelab.config import prf


class DownloaderBase(object):
    """

    Parameters
    ----------
    dt_fr : datetime.datetime
        The start datetime for downloading data.
    dt_to : datetime.datetime
        The end datetime for downloading data.
    root_dir_local : str or pathlib.Path
        The root directory in the local disk to store the downloaded data files.
    direct_download : bool
        Whether to start downloading once the Downloader object is created.
    force_download : bool
        Whether to force re-download even the data files are already in the local disk.
    dry_run : bool
        Whether to only print the downloading information without actual downloading.
    done : list of bool
        Whether the downloading is done. The length of the list is the number of files to be downloaded.
    file_paths_local : list of pathlib.Path
        The local file paths of the downloaded data files.
    file_paths_remote : list of str
        The remote file paths of the data files to be downloaded.
    
    """
    def __init__(
        self,
        dt_fr=None,
        dt_to=None,
        root_dir_local=None,
        direct_download=False,
        force_download=False,
        dry_run=False,
        download_from = None,
        **kwargs
    ):

        self.force_download = force_download
        self.dry_run = dry_run

        self.dt_fr = dt_fr
        self.dt_to = dt_to
        self.file_paths_source = []
        self.file_paths_local = []
        self.done = None
        self.file_paths_local = []
        self.file_paths_remote = []

        if root_dir_local is None:
            self.root_dir_local = prf.datahub_data_root_dir
        else:
            self.root_dir_local = root_dir_local

        if direct_download:
            self.download()
            
    def download(self, *args, **kwargs):
        pass
    
    def _to_download(self, file_path, with_suffix=None):
        
        to_download = True
        
        if with_suffix is not None:
            is_file = file_path.with_suffix(with_suffix).is_file()
        else:
            is_file = file_path.is_file()
        if is_file:
            if self.force_download:
                mylog.simpleinfo.info(
                    "The file {} exists in the directory {}: Forced redownloading the file ...".format(
                        file_path.name, file_path.parent.resolve()
                    )
                )
            else:
                mylog.simpleinfo.info(
                    "The file {} exists in the directory {}: Skipped downloading.".format(
                        file_path.name, file_path.parent.resolve()
                    )
                )
                to_download = False
        return to_download
    
    def _search_files(self, file_list, file_name_patterns):
        search_pattern = '.*' + '.*'.join(file_name_patterns) + '.*'
        fn_regex = re.compile(search_pattern)
        file_names = list(filter(fn_regex.match, file_list))
        return file_names
    


class DownloaderFromFTPBase(DownloaderBase):
    """
    Base downloader for downloading data files from FTP server.

    Parameters
    ----------
    ftp_host : str
        The FTP host address.
    ftp_port : int
        The FTP port [21].
    root_dir_remote : str
        The directory in the FTP that stores the data.
    """
    
    root_dir_remote = None

    def __init__(
        self,
        dt_fr, dt_to,
        ftp_host=None, ftp_port=21,
        username='anonymous', password='', 
        root_dir_local=None,
        root_dir_remote=None,
        direct_download=False, force_download=False, dry_run=False,
        **kwargs
        ):
        
        self.ftp_host = ftp_host
        self.ftp_port = ftp_port
        self.username = username
        self.password = password
        if root_dir_remote is not None:
            self.root_dir_remote = root_dir_remote
        else:
            self.root_dir_remote = '/'
        
        super(DownloaderFromFTPBase, self).__init__(
            dt_fr, dt_to,
            root_dir_local=root_dir_local, 
            direct_download=direct_download,
            force_download=force_download,
            dry_run=dry_run,
            **kwargs
        )
        
    def download(
        self, 
        with_TLS=False, 
        subdirs: None | list=None, 
        file_name_patterns: None | list=None,
        **kwargs
        ):
        
        timeout = kwargs.pop('timeout', 30) # seconds
        encoding = kwargs.pop('encoding', 'utf-8')
        
        if with_TLS:
            FTP_CLASS = ftplib.FTP_TLS
        else:
            FTP_CLASS = ftplib.FTP
            
        with FTP_CLASS(encoding=encoding) as ftp:
            ftp.connect(self.ftp_host, self.ftp_port, timeout)
            ftp.login(user=self.username, passwd=self.password)
            ftp.cwd(self.root_dir_remote)
            
            self.file_paths_remote = self.search_from_ftp(
                ftp, subdirs=subdirs, file_name_patterns=file_name_patterns
            )
            self.save_files_from_ftp(ftp)
            
    def search_from_ftp(self, ftp, subdirs: None | list=None, file_name_patterns: None | list=None):
        if subdirs is not None:
            for subdir in subdirs:
                ftp.cwd(subdir)
        try:
            files = ftp.nlst()
        except ftplib.error_perm as resp:
            if str(resp) == "550 No files found":
                mylog.StreamLogger.warning("No files in this directory.")
            else:
                raise

        file_names = self._search_files(
            file_list=files, file_name_patterns=file_name_patterns
        )        
        
        file_paths_remote = []
        for fn in file_names:
            file_paths_remote.append(ftp.pwd() + '/' + fn)
        return file_paths_remote
    
    def save_files_from_ftp(self, ftp, file_paths_local=None, root_dir_remote=None):
        if root_dir_remote is None:
            root_dir_remote = self.root_dir_remote
            
        if file_paths_local is None:
            self.file_paths_local = []
            for fp_remote in self.file_paths_remote:
                fp_local = self.root_dir_local
                for p in fp_remote.replace(root_dir_remote, '').split('/'):
                    fp_local = fp_local / p
                self.file_paths_local.append(
                    fp_local
                )
        self.done = [False] * len(self.file_paths_remote)
        
        if self.dry_run:
            for i, (fp_remote, fp_local) in enumerate(zip(self.file_paths_remote, self.file_paths_local)):
                mylog.simpleinfo.info(
                    f"Dry run: Downloading the file {fp_remote} to {fp_local} ..."
                )
            return
        
        for i, (fp_remote, fp_local) in enumerate(zip(self.file_paths_remote, self.file_paths_local)):
            to_download = self._to_download(fp_local)
            
            if not to_download:
                self.done[i] = True
                continue
            mylog.simpleinfo.info(
                f"Downloading the file: {fp_remote} ..."
            )
            
            file_dir_remote = fp_remote.rsplit('/', 1)[0]
            file_name_remote = fp_remote.split('/')[-1]
            
            file_dir_local = fp_local.parent.resolve()
            file_dir_local.mkdir(exist_ok=True, parents=True)
                
            with open(fp_local, 'w+b') as f:
                if file_dir_remote != ftp.pwd():
                    ftp.cwd(file_dir_remote)
                
                bufsize=1024
                total=ftp.size(fp_remote)
                pbar=tqdm(total=total)
                def bar(data):
                    f.write(data)
                    pbar.update(len(data))
                res = ftp.retrbinary('RETR '+ file_name_remote, bar, bufsize)
                pbar.close()
                # res = ftp.retrbinary('RETR ' + file_name_remote, f.write)
                # mylog.simpleinfo.info(res)
                if not res.startswith('226'):
                    mylog.StreamLogger.warning('The downloading is not compiled...: {}'.format(res))
                    fp_local.unlink(missing_ok=True)
                self.done[i] = True
                mylog.simpleinfo.info("Saved as {}.".format(fp_local))
            if not self.done[i]:
                mylog.StreamLogger.error("Error in downloading the file {}.".format(file_name_remote))
        return


class DownloaderFromHTTPBase(DownloaderBase):
    """
    Base downloader for downloading data files from HTTP server.

    Parameters
    ----------
    base_url : str
        The base URL of the HTTP server.
    root_dir_remote : str
        The directory in the HTTP server that stores the data.
    """
    
    root_dir_remote = None

    def __init__(
        self,
        dt_fr, dt_to,
        base_url=None,
        root_dir_local=None,
        root_dir_remote=None,
        direct_download=False, force_download=False, dry_run=False,
        **kwargs
        ):
        
        self.base_url = base_url
        
        if root_dir_remote is not None:    
            self.root_dir_remote = root_dir_remote
        
        super(DownloaderFromHTTPBase, self).__init__(
            dt_fr, dt_to,
            root_dir_local=root_dir_local,
            direct_download=direct_download,
            force_download=force_download,
            dry_run=dry_run,
            **kwargs
        )
        
    def download(
        self, 
        subdirs: None | list=None, 
        file_name_patterns: None | list=None,
        **kwargs
        ):
        
        self.file_paths_remote = self.search_from_http(
            subdirs=subdirs, file_name_patterns=file_name_patterns
        )
        self.save_files_from_http()
        
        
    def search_from_http(self, subdirs: None | list=None, file_name_patterns: None | list=None, **kwargs):
        url_patterns = [self.base_url]
        if str(self.root_dir_remote):
            url_patterns.append(self.root_dir_remote)
        if subdirs is not None:
            url_patterns.extend(subdirs)
        url = '/'.join(url_patterns)
        r = requests.get(url)
        soup = bs4.BeautifulSoup(r.text, 'html.parser')
        a_tags = soup.find_all('a', href=True)
        hrefs = [a_tag['href'] for a_tag in a_tags]
        
        hrefs = self._search_files(file_list=hrefs, file_name_patterns=file_name_patterns)

        file_paths_remote = []
        for href in hrefs:
            file_paths_remote.append(url + '/' + href)
        r.close()
        return file_paths_remote

    def save_files_from_http(self, file_paths_local=None, root_dir_remote=None):
        if root_dir_remote is None:
            root_dir_remote = self.root_dir_remote
        if str(root_dir_remote):
            root_url = self.base_url + '/' + root_dir_remote
        else:
            root_url = self.base_url
            
        if file_paths_local is None:
            self.file_paths_local = []
            for fp_remote in self.file_paths_remote:
                fp_local = self.root_dir_local
                for p in fp_remote.replace(root_url, '').split('/'):
                    fp_local = fp_local / p
                self.file_paths_local.append(
                    fp_local
                )
        else:
            self.file_paths_local = file_paths_local
        self.done = [False] * len(self.file_paths_remote)
        
        if self.dry_run:
            for i, (fp_remote, fp_local) in enumerate(zip(self.file_paths_remote, self.file_paths_local)):
                mylog.simpleinfo.info(
                    f"Dry run: Downloading the file {fp_remote} to {fp_local} ..."
                )
            return
        
        for i, (fp_remote, fp_local) in enumerate(zip(self.file_paths_remote, self.file_paths_local)):
            to_download = self._to_download(fp_local)
            
            if not to_download:
                self.done[i] = True
                continue
            
            mylog.simpleinfo.info(
                f"Downloading the file: {fp_remote} ..."
            )

            res = self._download_by_requests_get(fp_remote, fp_local)
            if res:
                mylog.simpleinfo.info(f"Saved as {fp_local}.")
                self.done[i] = True

        return

    @staticmethod
    def _download_by_requests_get(
        url, file_path_local, 
        params=None, stream=True, allow_redirects=True, 
        file_block_size=1024, 
        **kwargs
        ):
        try: 
            r = requests.get(url, params=params, stream=stream, allow_redirects=allow_redirects, **kwargs)
            if r.status_code != 200:
                return -2

            total_size_in_bytes = int(r.headers.get('content-length', 0))

            progress_bar = tqdm.tqdm(total=total_size_in_bytes, unit='iB', unit_scale=True)

            file_path_local.parent.mkdir(parents=True, exist_ok=True)
            with open(file_path_local, 'wb') as file:
                for data in r.iter_content(file_block_size):
                    progress_bar.update(len(data))
                    file.write(data)
            progress_bar.close()
            if total_size_in_bytes != 0 and progress_bar.n != total_size_in_bytes:
                mylog.StreamLogger.error("Something wrong during the download!")
                file_path_local.unlink(missing_ok=True)
                return -1
            r.close()
        except Exception as e:
            mylog.StreamLogger.error(f"Exception during downloading the file from {url}: {e}")
            return -1
        return 1


class CDAWebHTTPDownloader(DownloaderFromHTTPBase):
    """
    Downloader for downloading data files from CDAWeb HTTP server.

    Parameters
    ----------
    base_url : str
        The base URL of the CDAWeb HTTP server.
    root_dir_remote : str
        The directory in the CDAWeb HTTP server that stores the data.
    """

    root_dir_remote = 'pub/data'
    
    def __init__(
        self,
        dt_fr, dt_to,
        root_dir_local=None,
        root_dir_remote=None,
        direct_download=False, force_download=False, dry_run=False,
        **kwargs
        ):
        
        base_url = 'https://cdaweb.gsfc.nasa.gov'
        
        if root_dir_local is None:
            root_dir_local = prf.datahub_data_root_dir / 'CDAWeb'
        
        super().__init__(
            dt_fr, dt_to,
            base_url=base_url,
            root_dir_local=root_dir_local,
            root_dir_remote=root_dir_remote,
            direct_download=direct_download,
            force_download=force_download,
            dry_run=dry_run,
            **kwargs
        )


import geospacelab.datahub.sources.madrigal.utilities as madrigal
import geospacelab.datahub.sources.madrigal.isr.eiscat as eiscat

fn = "EISCAT_2021-03-10_beata_ant@uhfa.hdf5"
madrigal.show_hdf5_structure(filename=fn, filepath="")
madrigal.show_hdf5_metadata(filename=fn, filepath="")
eiscat.list_eiscat_hdf5_variables("./examples/" + fn)

fn = "MAD6300_2021-03-10_beata_ant@uhfa.hdf5"
madrigal.show_hdf5_structure(filename=fn, filepath="")
madrigal.show_hdf5_metadata(filename=fn, filepath="")

fn = "MAD6400_2021-03-10_beata_ant@uhfa.hdf5"
madrigal.show_hdf5_structure(filename=fn, filepath="")
madrigal.show_hdf5_metadata(filename=fn, filepath="")



# Licensed under the BSD 3-Clause License
# Copyright (C) 2021 GeospaceLab (geospacelab)
# Author: Lei Cai, Space Physics and Astronomy, University of Oulu

__author__ = "Lei Cai"
__copyright__ = "Copyright 2021, GeospaceLab"
__license__ = "BSD-3-Clause License"
__email__ = "lei.cai@oulu.fi"
__docformat__ = "reStructureText"


"""
To download the EISCAT quickplots and analyzed data archived in http://portal.eiscat.se/schedule/schedule.cgi
By Lei Cai on 2021.04.01
"""
import datetime
from dateutil.relativedelta import relativedelta
from dateutil.parser import parse as dtparse
import numpy as np
import re
import requests
import bs4
import os
import pathlib

from geospacelab.config import pref as pfr
import geospacelab.datahub.sources.madrigal as madrigal
import geospacelab.toolbox.utilities.pylogging as mylog
import geospacelab.toolbox.utilities.pydatetime as dttool

from geospacelab.datahub.sources.madrigal.downloader import Downloader as DownloaderBase


class Downloader(DownloaderBase):

    def __init__(
            self, 
            dt_fr: datetime.datetime, dt_to: datetime,
            antennas=None, 
            data_file_root_dir=None, 
            kind_data="madrigal",
            user_fullname=madrigal.default_user_fullname,
            user_email=madrigal.default_user_email,
            user_affiliation=madrigal.default_user_affiliation,
            include_exp_name_patterns: list=None,
            exclude_exp_name_patterns: list=None,
            include_exp_ids: list=None,
            exclude_exp_ids: list=None,
            include_file_name_patterns: list = None,
            exclude_file_name_patterns: list = None,
            include_file_type_patterns=None,
            exclude_file_type_patterns=None,
            direct_download = True,
            force_download = False,
            dry_run: bool=False,
            madrigal_url: str = "http://madrigal.eiscat.se/",
        ):

        # dt_fr = dttool.get_start_of_the_day(dt_fr)
        # dt_to = dttool.get_end_of_the_day(dt_to)
        
        icodes = []
        for a in antennas:
            icodes.extend(instrument_codes[a])  
        icodes = [int(i) for i in np.unique(icodes)]
        
        self.kind_data=kind_data

        if data_file_root_dir is None:
            self.data_file_root_dir = pfr.datahub_data_root_dir / 'Madrigal' / 'EISCAT' / 'Analyzed'
        else:
            self.data_file_root_dir = data_file_root_dir

        super().__init__(
            dt_fr=dt_fr, dt_to=dt_to, icodes=icodes,
            include_exp_name_patterns=include_exp_name_patterns, 
            exclude_exp_name_patterns=exclude_exp_name_patterns, 
            include_exp_ids=include_exp_ids,
            exclude_exp_ids=exclude_exp_ids,
            include_file_name_patterns=include_file_name_patterns, 
            exclude_file_name_patterns=exclude_file_name_patterns,
            include_file_type_patterns=include_file_type_patterns, 
            exclude_file_type_patterns=exclude_file_type_patterns,  
            data_file_root_dir=data_file_root_dir,
            force_download=force_download, direct_download=direct_download, dry_run=dry_run,
            madrigal_url=madrigal_url,
            user_fullname=user_fullname, user_email=user_email, user_affiliation=user_affiliation)

    def download(self, **kwargs):
        if self.kind_data.lower() == 'eiscat':
            self.download_EISCAT()
        elif self.kind_data.lower() == 'madrigal':
            self.download_MAD()
        else:
            raise NotImplementedError    
        return 
    
    def download_EISCAT(self):
        exps, database = self.get_exp_list(
            dt_fr=dttool.get_start_of_the_day(self.dt_fr),
            dt_to=dttool.get_end_of_the_day(self.dt_to),
            include_exp_name_patterns=self.include_exp_name_patterns,
            exclude_exp_name_patterns=self.exclude_exp_name_patterns,
            include_exp_ids=self.include_exp_ids,
            exclude_exp_ids=self.exclude_exp_ids,
            icodes=self.icodes,
            madrigal_url=self.madrigal_url,
            display=True)
        self.exp_list = list(exps)
        self.database = database
        
        cookies = {
                'user_email': self.user_email,
                'user_fullname': self.user_fullname,
                'user_affiliation': self.user_affiliation
            }
        for exp in exps:
            dt_fr_exp = datetime.datetime(
                exp.startyear, exp.startmonth, exp.startday, exp.starthour, exp.startmin, exp.startsec
            )
            dt_to_exp = datetime.datetime(
                exp.endyear, exp.endmonth, exp.endday, exp.endhour, exp.endmin, exp.endsec
            )
            if (dt_fr_exp >= self.dt_to) or (dt_to_exp <= self.dt_fr):
                continue
            try:
                res = re.search(r'([\d]{4}\-[\d]{2}\-[\d]{2})_(\w+)@(\w+)', exp.name)
                thisday = datetime.datetime.strptime(res.groups()[0], "%Y-%m-%d")
                pulse_code = res.groups()[1]
                antenna_ = res.groups()[2]
                if 'uhf' in antenna_.lower():
                    antenna = 'UHF'
                elif 'vhf' in antenna_.lower():
                    antenna = 'VHF'
                elif any([a in antenna_.lower() for a in ['32m', '42m', 'esr']]):
                    antenna = 'ESR'
                elif 'sod' in antenna_.lower():
                    antenna = 'SOD'
                elif 'kir' in antenna_.lower():
                    antenna = 'KIR'
                else:
                    raise NotImplementedError
            except Exception as e:
                print(e)
                mylog.StreamLogger.warning("Parsing the experiment name was failed! EXP: {}".format(exp.name))
            
            url = "https://madrigal.eiscat.se/madrigal/showExperiment?experiment_list=" + str(exp.id) + "&show_plots="
            
            r = requests.get(url, cookies=cookies)
            soup = bs4.BeautifulSoup(r.text, 'html.parser')
            links = soup.find_all('a', href=True)
            for link in links:
                href = link['href']
                if any(href.endswith(s) for s in ['.png', '.tar.gz', '.hdf5']):
                    filename = href.split('/')[-1]
                    file_dir_local = self.data_file_root_dir / antenna / thisday.strftime("%Y") / \
                                (exp.name + '_EID-' + str(exp.id) + '_'
                                 + dt_fr_exp.strftime("%Y%m%dT%H%M%S") + '_'
                                 + dt_to_exp.strftime("%Y%m%dT%H%M%S"))
                    file_dir_local.mkdir(parents=True, exist_ok=True)

                    remote_file = requests.get(href)
                    file_path = file_dir_local / filename
                    if file_path.is_file() and not self.force_download:
                        print("The file {} has been downloaded.".format(filename))
                        self.done=True
                        continue
                    mylog.simpleinfo.info(
                        'Downloading "{} ..."'.format(filename)
                    )
                    with open(file_path, "wb") as eiscat:
                        eiscat.write(remote_file.content)
                    mylog.simpleinfo.info('Saved to {}.'.format(file_dir_local))
                    self.done = True
        return 
    
    def download_MAD(self, **kwargs):

        exps, database = self.get_exp_list(
            dt_fr=dttool.get_start_of_the_day(self.dt_fr),
            dt_to=dttool.get_end_of_the_day(self.dt_to),
            include_exp_name_patterns=self.include_exp_name_patterns,
            exclude_exp_name_patterns=self.exclude_exp_name_patterns,
            include_exp_ids=self.include_exp_ids,
            exclude_exp_ids=self.exclude_exp_ids,
            icodes=self.icodes,
            madrigal_url=self.madrigal_url,
            display=True)
        self.exp_list = list(exps)
        self.database = database

        exps, exps_error = self.get_online_file_list(
            exp_list=self.exp_list, database=database,
            include_file_name_patterns=self.include_file_name_patterns,
            exclude_file_name_patterns=self.exclude_file_name_patterns,
            include_file_type_patterns=self.include_file_type_patterns,
            exclude_file_type_patterns=self.exclude_file_type_patterns,
            display=True
        )
        self.exp_list_error = list(exps_error)

        file_paths = [] 
        for exp in exps:
            dt_fr_exp = datetime.datetime(
                exp.startyear, exp.startmonth, exp.startday, exp.starthour, exp.startmin, exp.startsec
            )
            dt_to_exp = datetime.datetime(
                exp.endyear, exp.endmonth, exp.endday, exp.endhour, exp.endmin, exp.endsec
            )
            # if (dt_fr_exp >= self.dt_to) or (dt_to_exp <= self.dt_fr):
            #     continue
            for file in list(exp.files):

                file_path_remote = pathlib.Path(file.name)
                file_name_remote = file_path_remote.name

                try:
                    res = re.search(r'([\d]{4}\-[\d]{2}\-[\d]{2})_(\w+)@(\w+)', exp.name)
                    thisday = datetime.datetime.strptime(res.groups()[0], "%Y-%m-%d")
                    pulse_code = res.groups()[1]
                    antenna_ = res.groups()[2]
                    if 'uhf' in antenna_.lower():
                        antenna = 'UHF'
                    elif 'vhf' in antenna_.lower():
                        antenna = 'VHF'
                    elif any([a in antenna_.lower() for a in ['32m', '42m', 'esr']]):
                        antenna = 'ESR'
                    elif 'sod' in antenna_.lower():
                        antenna = 'SOD'
                    elif 'kir' in antenna_.lower():
                        antenna = 'KIR'
                    else:
                        raise NotImplementedError
                except Exception as e:
                    print(e)
                    mylog.StreamLogger.warning("Parsing the experiment name was failed! EXP: {}".format(exp.name))

                file_dir_local = self.data_file_root_dir / antenna / thisday.strftime("%Y") / \
                                (exp.name + '_EID-' + str(exp.id) + '_'
                                 + dt_fr_exp.strftime("%Y%m%dT%H%M%S") + '_'
                                 + dt_to_exp.strftime("%Y%m%dT%H%M%S"))
                file_dir_local.mkdir(parents=True, exist_ok=True)

                file_name_local = file_name_remote
                file_path_local = file_dir_local / file_name_local

                super().download(
                    file_path_remote=file.name, file_path_local=file_path_local,
                    file_format='hdf5')
                file_paths.append(file_path_local)
        return file_paths


def test():
    dt_0 = datetime.datetime(2016, 3, 9, 19)
    dt_1 = datetime.datetime(2016, 3, 9, 23, 59)
    downloader = DownloaderNew(
        dt_fr=dt_0, dt_to=dt_1, 
        kind_data='EISCAT',
        antennas=['UHF', 'VHF'],
        force_download=True,
        exclude_file_type_patterns=['pp']
        )
    pass

class DownloaderOld(object):
    """Download the quickplots and archieved analyzed results from EISCAT schedule webpage
    """

    def __init__(self, dt_fr, dt_to, sites=None, data_file_root_dir=None, kind_data="EISCAT", download_pp=False,
                 user_fullname=madrigal.default_user_fullname,
                 user_email=madrigal.default_user_email,
                 user_affiliation=madrigal.default_user_affiliation):
        self.user_fullname = user_fullname
        self.user_email = user_email
        self.user_affiliation = user_affiliation

        dt_fr = dttool.get_start_of_the_day(dt_fr)
        dt_to = dttool.get_end_of_the_day(dt_to)
        if dt_fr == dt_to:
            dt_to = dt_to + datetime.timedelta(hours=23, minutes=59)
        self.dt_fr = dt_fr  # datetime from
        self.dt_to = dt_to  # datetime to
        if sites is None:
            sites = ['UHF', 'VHF', 'KIR', 'SOD',
                     'ESR']  # site list: default is the list of all the sites, use e.g., ['UHF'] for single site
        self.sites = sites

        if data_file_root_dir is None:
            self.data_file_root_dir = pfr.datahub_data_root_dir / 'Madrigal' / 'EISCAT' / 'analyzed'
        else:
            self.data_file_root_dir = data_file_root_dir
        self.done = False

        if kind_data.lower() == "madrigal":
            self.madrigal_url = "https://madrigal.eiscat.se/madrigal/"
            self.download_madrigal_files(download_pp=download_pp)
        elif kind_data.lower() == "eiscat":
            self.url_base = 'http://portal.eiscat.se'  # EISCAT schedule webpage
            self.urls = []
            self.search_scheduled_url()
            self.download_eiscat_files()

    def search_scheduled_url(self):
        """
        Search available urls
        :return: None
        """
        diff_month = (self.dt_to.year - self.dt_fr.year) * 12 + self.dt_to.month - self.dt_fr.month
        urls = set()
        for nm in range(diff_month + 1):
            dt_now = datetime.datetime(self.dt_fr.year, self.dt_fr.month, 1) + relativedelta(months=nm)

            payload_options = []
            for site in self.sites:
                payload_options.append(site)
                if site == 'UHF':
                    payload_options.append('TRO')  # full list: VHF=on&UHF=on&TRO=on&KIR=on&SOD=on&ESR=on&HEA=on
            payload = {
                'year': "{:04d}".format(dt_now.year),
                'month': '{:d}'.format(dt_now.month),
                'A': 'on'
            }
            for opt in payload_options:
                payload[opt] = 'on'

            r = requests.get(self.url_base + '/schedule/schedule.cgi', params=payload)
            soup = bs4.BeautifulSoup(r.text, 'html.parser')

            strong_tags = soup.find_all('strong')

            for stag in strong_tags:
                yymm = dt_now.strftime("%Y:%m")
                if yymm not in stag.string:
                    continue
                tagline = stag.sourceline
                # print(stag.string)
                content = r.text.split('\n')[tagline - 1]
                # print(content)
                soup_line = bs4.BeautifulSoup(content, 'html.parser')
                link = soup_line.find('a', href=True)['href']
                if 'experiment_list' not in link:
                    print('No link for the existing experiment ({})'.format(link))
                    continue
                urls.add(link)
            self.urls = urls
        return None

    def download_eiscat_files(self):
        """
        Download the files
        """
        if not list(self.urls):
            mylog.StreamLogger.info("No experiments available!")
            return None

        for url in self.urls:
            cookies = {
                'user_email': self.user_email,
                'user_fullname': self.user_fullname,
                'user_affiliation': self.user_affiliation
            }
            r = requests.get(url, cookies=cookies)
            soup = bs4.BeautifulSoup(r.text, 'html.parser')
            links = soup.find_all('a', href=True)
            for link in links:
                href = link['href']
                if any(href.endswith(s) for s in ['.png', '.tar.gz', '.hdf5']):
                    filename = href.split('/')[-1]

                    match = re.search(r'\d{4}-\d{2}-\d{2}', filename)
                    thisday = datetime.datetime.strptime(match.group(), '%Y-%m-%d')
                    if thisday < self.dt_fr or thisday > self.dt_to:
                        continue
                    for key, value in locs.items():
                        if '/' + value + '/' in href:
                            site = key

                    #site1 = re.search("[@|_][a-zA-Z0-9]*[.]", filename).group(0)[1:-1]
                    search_pattern = re.search(r"\d{4}-\d{2}-\d{2}_[a-zA-Z0-9]*", filename).group(0)
                    sub_dir = search_pattern + '@' + site
                    data_file_dir = self.data_file_root_dir / site / thisday.strftime('%Y') / sub_dir
                    data_file_dir.mkdir(parents=True, exist_ok=True)

                    remote_file = requests.get(href)
                    file_path = data_file_dir / filename
                    if file_path.is_file():
                        print("The file {} has been downloaded.".format(filename))
                        continue
                    mylog.simpleinfo.info(
                        'Downloading "{} from portal.eiscat.se to {} ..."'.format(filename, data_file_dir)
                    )
                    with open(file_path, "wb") as eiscat:
                        eiscat.write(remote_file.content)
                    mylog.simpleinfo.info('Done!')
                    self.done = True
        return

    def download_madrigal_files(self, download_pp=False):
        icodes = []
        for site in self.sites:
            icodes.extend(instrument_codes[site])
        for icode in icodes:
            exp_list, _, database = madrigal.utilities.list_experiments(
                icode, self.dt_fr, self.dt_to, madrigal_url=self.madrigal_url)
            for exp in exp_list:
                files = database.getExperimentFiles(exp.id)
                for file in files:
                    if not download_pp and 'GUISDAP pp' in file.kindatdesc:
                        continue
                    file_path = pathlib.Path(file.name)
                    site = file_path.name.split("@")[1][0:3].upper()
                    if '32' in site or '42' in site:
                        site = 'ESR'

                    match = re.search(r'\d{4}-\d{2}-\d{2}', file_path.name)
                    dt_str = match.group(0)
                    thisday = datetime.datetime.strptime(dt_str, "%Y-%m-%d")
                    if thisday < self.dt_fr or thisday > self.dt_to:
                        continue

                    # sub_dir = file_path.name.split('_', maxsplit=1)[1]
                    search_pattern = re.search(r"\d{4}-\d{2}-\d{2}_[a-zA-Z0-9]*", file_path.name).group(0)
                    sub_dir = search_pattern + '@' + site
                    data_file_dir = self.data_file_root_dir / site / dt_str[0:4] / sub_dir
                    data_file_dir.mkdir(parents=True, exist_ok=True)
                    data_file_path = data_file_dir / file_path.name
                    if data_file_path.is_file():
                        mylog.simpleinfo.info("The file {} has been downloaded.".format(data_file_path.name))
                        continue

                    mylog.simpleinfo.info("Downloading  {} from the Madrigal database ...".format(file_path.name))
                    database.downloadFile(
                        file.name, data_file_path,
                        self.user_fullname, self.user_email, self.user_affiliation,
                        "hdf5"
                    )
                    self.done = True
                    mylog.simpleinfo.info("Done!")


        # fhdf5 = h5py.File(outDir + fn, 'r')


instrument_codes = {
    'UHF': [72],
    'VHF': [74],
    '42m': [95],
    '32m': [95],
    'SOD': [73, 76],
    'KIR': [71, 75]
}

locs = {
    'UHF': 'tro',
    'VHF': 'eis',
    'ESR': 'lyr',
    'SOD': 'sod',
    'KIR': 'kir'
}


class EISCATSchedule(object):
    """ Analyze EISCAT schedule
    """

    def __init__(self, dt_fr, dt_to, sites=None, root_filepath=None, monthly=False):
        self.url_base = 'http://portal.eiscat.se'  # EISCAT schedule webpage
        self.dt_fr = dt_fr
        self.dt_to = dt_to  # datetime
        if sites is None:
            sites = ['UHF', 'VHF', 'KIR', 'SOD',
                     'ESR']  # site list: default is the list of all the sites, use e.g., ['UHF'] for single site
        self.sites = sites
        self.rs = []
        self.experiments = []

        self.urls = []
        if root_filepath is None:
            root_filepath = pfr.datahub_data_root_dir  # root path for the download files. Default: the current work directory + /results/
        self.root_filepath = root_filepath
        self.monthly = monthly  # True: download the whole month data. False: only for the specific date
        self.search_archives()
        self.analyze_archives()
        self.to_txt()

    def search_archives(self):
        """
        Search available urls
        :return: None
        """

        diff_month = (self.dt_to.year - self.dt_fr.year) * 12 + self.dt_to.month - self.dt_fr.month
        for nm in range(diff_month + 1):
            dt_now = datetime.datetime(self.dt_fr.year, self.dt_fr.month, 1) + relativedelta(months=nm)

            payload_options = []
            for site in self.sites:
                payload_options.append(site)
                if site == 'UHF':
                    payload_options.append('TRO')  # full list: VHF=on&UHF=on&TRO=on&KIR=on&SOD=on&ESR=on&HEA=on
            payload = {
                'year': "{:04d}".format(dt_now.year),
                'month': '{:d}'.format(dt_now.month),
                'A': 'on'
            }
            for opt in payload_options:
                payload[opt] = 'on'

            r = requests.get(self.url_base + '/schedule/schedule.cgi', params=payload)

            self.rs.append(r)

    def analyze_archives(self):
        nexp = 0
        for r in self.rs:
            soup = bs4.BeautifulSoup(r.text, 'html.parser')
            schedule = soup.pre.text.split('\n')
            for item_text in schedule:
                if not is_date(item_text[0:10].replace(':', '-')):
                    continue
                dt_now = datetime.datetime.strptime(item_text[0:10], '%Y:%m:%d')
                if not self.monthly:
                    if dt_now < self.dt_fr or dt_now > self.dt_to:
                        continue

                time_rec = item_text[15:63]

                inds = [x for x, v in enumerate(time_rec) if v == 'A']
                if not list(inds):
                    continue

                info = item_text[65:-1].split(' ')
                # print(dt_fr)
                # print(info)
                antenna = info[0]
                any_pattern = False
                for site in self.sites:
                    if site.lower() in antenna.lower():
                        any_pattern = True
                    if site == 'ESR':
                        if ('32' in antenna) or ('42' in antenna):
                            any_pattern = True
                if not any_pattern:
                    continue

                exp = info[1]
                info = item_text[65:-1].split(')')[-1].strip()

                info_splits = info.split('_')
                code = info_splits[0]
                mode = ''
                version = ''
                if len(info_splits) > 1:
                    mode = info_splits[1]
                if len(info_splits) > 2:
                    version = info_splits[2]

                # if not (('ip2' in mode.lower()) or 'cp2' in mode.lower()):
                #    continue
                # try:
                #     code = info.split('_')[0]
                #     mode = info.split('_')[1]
                #     version = info.split('_')[2]
                # except IndexError:
                #     print('Not a standard experiment {}'.format(info))
                #     continue

                ind_diff = np.diff(inds)
                ind_gaps = np.where(ind_diff > 3)[0]  # time difference greater than 1.5 h
                ind_ranges = []
                for ind_, ind_gap in enumerate(ind_gaps):
                    if ind_ == 0:
                        ind_fr = inds[0]
                        ind_to = inds[ind_gap]
                    else:
                        ind_fr = inds[ind_gaps[ind_ - 1] + 1]
                        ind_to = inds[ind_gap]
                    ind_ranges.append([ind_fr, ind_to])

                if list(ind_gaps):
                    ind_ranges.append([inds[ind_gaps[-1] + 1], inds[-1]])
                else:
                    ind_ranges.append([inds[0], inds[-1]])

                for ind_range in ind_ranges:
                    ind_1 = ind_range[0]
                    ind_2 = ind_range[1]
                    dt_fr = dt_now + datetime.timedelta(hours=ind_1 / 2)
                    HH = (ind_2 + 1) / 2
                    if HH >= 24:
                        HH = 23
                        MM = 59
                    else:
                        MM = 0
                    dt_to = dt_now + datetime.timedelta(hours=HH, minutes=MM)
                    suffix = 'OK'
                    if time_rec.count('A') < (ind_2 - ind_1 + 1):
                        suffix = '?'

                    self.experiments.append([dt_fr, dt_to, antenna, code, mode, version, exp, suffix, time_rec])
                    nexp = nexp + 1

        if nexp == 0:
            print('No available archived experiments!')
            return False

        return True

    def to_txt(self, file_name=None, file_path=None):
        if file_name is None:
            file_name = "EISCAT_archives.txt"
        if file_path is None:
            file_path = self.root_filepath

        with open(os.path.join(file_path, file_name), 'w') as out_file:
            for line in self.experiments:
                line_str = "{0:<17s}{1:<7s}{2:<5s}{3:<8s}{4:<30s}{5:<40s}{6:<40s}{7:<80s}{8:<65s}".format(
                    line[0].strftime('%Y-%m-%d %H:%M'), line[1].strftime('%H:%M'),
                    line[7], line[2], line[3], line[4], line[5], line[6], line[8]) + '\n'
                out_file.write(line_str)


def is_date(string, fuzzy=False):
    """
    Return whether the string can be interpreted as a date.

    :param string: str, string to check for date
    :param fuzzy: bool, ignore unknown tokens in string if True
    """
    try:
        dtparse(string, fuzzy=fuzzy)
        return True

    except ValueError:
        return False


if __name__ == "__main__":
    test()



# Licensed under the BSD 3-Clause License
# Copyright (C) 2021 GeospaceLab (geospacelab)
# Author: Lei Cai, Space Physics and Astronomy, University of Oulu

import datetime

import numpy as np

import geospacelab.datahub as datahub
from geospacelab.datahub import DatabaseModel, FacilityModel, InstrumentModel, ProductModel
from geospacelab.datahub.sources.jhuapl import jhuapl_database
from geospacelab.config import prf
import geospacelab.toolbox.utilities.pydatetime as dttool
import geospacelab.toolbox.utilities.pybasic as basic
import geospacelab.toolbox.utilities.pylogging as mylog
from geospacelab.datahub.sources.cdaweb.dmsp.ssusi.sdr_disk.loader import Loader as default_Loader
from geospacelab.datahub.sources.cdaweb.dmsp.ssusi.sdr_disk.downloader import Downloader as default_Downloader
import geospacelab.datahub.sources.cdaweb.dmsp.ssusi.sdr_disk.variable_config as var_config


default_dataset_attrs = {
    'database': jhuapl_database,
    'facility': 'DMSP',
    'instrument': 'SSUSI',
    'product': 'SDR_DISK',
    'data_file_ext': 'nc',
    'data_root_dir': prf.datahub_data_root_dir / 'CDAWeb' / 'DMSP' / 'SSUSI' / 'SDR_DISK',
    'allow_load': True,
    'allow_download': True,
    'data_search_recursive': False,
    'label_fields': ['database', 'facility', 'instrument', 'product'],
    'time_clip': False,
}

default_variable_names = [
    'STARTING_TIME', 'STOPPING_TIME', 'DATETIME',
    'FILE_VERSION', 'DATA_PRODUCT_VERSION', 'SOFTWARE_VERSION_NUMBER', 'CALIBRATION_PERIOD_VERSION', 'EMISSION_SPECTRA',
    'SC_DATETIME', 'SC_ORBIT_ID', 'SC_GEO_LAT', 'SC_GEO_LON', 'SC_GEO_ALT',
    'DISK_GEO_LAT', 'DISK_GEO_LON', 'DISK_GEO_ALT', 'DISK_SZA', 'DISK_SAA',
    'ACROSS_PIXEL_SIZE', 'ALONG_PIXEL_SIZE', 'EFFECTIVE_LOOK_ANGLE', 'EXPOSURE', 'SAA_COUNT', 'DARK_COUNT_CORRECTION',
    'SCATTER_LIGHT_1216_CORRECTION', 'SCATTER_LIGHT_1304_CORRECTION', 'OVERLAP_1304_1356_CORRECTION', 'LONG_WAVE_SCATTER_CORRECTION',
    'RED_LEAK_CORRECTION', 'DQI',
    'DISK_COUNTS_1216', 'DISK_COUNTS_1304', 'DISK_COUNTS_1356', 'DISK_COUNTS_LBHS', 'DISK_COUNTS_LBHL',
    'DISK_R_1216', 'DISK_R_1304', 'DISK_R_1356', 'DISK_R_LBHS', 'DISK_R_LBHL',
    'DISK_R_RECT_1216', 'DISK_R_RECT_1304', 'DISK_R_RECT_1356', 'DISK_R_RECT_LBHS', 'DISK_R_RECT_LBHL',
    'DISK_R_1216_ERROR', 'DISK_R_1304_ERROR', 'DISK_R_1356_ERROR', 'DISK_R_LBHS_ERROR', 'DISK_R_LBHL_ERROR',
    'DISK_R_RECT_1216_ERROR', 'DISK_R_RECT_1304_ERROR', 'DISK_R_RECT_1356_ERROR', 'DISK_R_RECT_LBHS_ERROR', 'DISK_R_RECT_LBHL_ERROR',
    'DISK_DECOMP_1216_ERROR', 'DISK_DECOMP_1304_ERROR', 'DISK_DECOMP_1356_ERROR', 'DISK_DECOMP_LBHS_ERROR', 'DISK_DECOMP_LBHL_ERROR',
    'DISK_CALIB_1216_ERROR', 'DISK_CALIB_1304_ERROR', 'DISK_CALIB_1356_ERROR', 'DISK_CALIB_LBHS_ERROR', 'DISK_CALIB_LBHL_ERROR',
    'DQI_1216', 'DQI_1304', 'DQI_1356', 'DQI_LBHS', 'DQI_LBHL'
]

# default_data_search_recursive = True

default_attrs_required = ['sat_id', 'orbit_id', 'pp_type', 'pole']


class Dataset(datahub.DatasetSourced):
    def __init__(self, **kwargs):
        kwargs = basic.dict_set_default(kwargs, **default_dataset_attrs)

        super().__init__(**kwargs)

        self.database = kwargs.pop('database', 'JHUAPL')
        self.facility = kwargs.pop('facility', 'DMSP')
        self.instrument = kwargs.pop('instrument', 'SSUSI')
        self.product = kwargs.pop('product', 'SDR_DISK')
        self.allow_download = kwargs.pop('allow_download', True)
        self.force_download = kwargs.pop('force_download', False)
        self.download_dry_run = kwargs.pop('download_dry_run', False)

        self.sat_id = kwargs.pop('sat_id', '')
        self.orbit_id = kwargs.pop('orbit_id', None)
        self.pole = kwargs.pop('pole', '')
        self.pp_type = kwargs.pop('pp_type', '')

        self.metadata = None

        allow_load = kwargs.pop('allow_load', False)

        # self.config(**kwargs)

        if self.loader is None:
            self.loader = default_Loader

        if self.downloader is None:
            self.downloader = default_Downloader

        self._validate_attrs()

        if allow_load:
            self.load_data()

    def _validate_attrs(self):
        for attr_name in default_attrs_required:
            attr = getattr(self, attr_name)
            if not str(attr):
                mylog.StreamLogger.warning("The parameter {} is required before loading data!".format(attr_name))
            if attr_name == 'orbit_id':
                if attr is None or attr == '':
                    mylog.StreamLogger.warning("For a fast process, it's better to specify the orbit id.")

    def label(self, **kwargs):
        label = super().label()
        return label

    def load_data(self, **kwargs):
        self.check_data_files(**kwargs)

        self._set_default_variables(
            default_variable_names,
            configured_variables=var_config.configured_variables
        )
        for file_path in self.data_file_paths:
            load_obj = self.loader(file_path, file_type=self.product.lower(), pole=self.pole, pp_type=self.pp_type)

            for var_name in self._variables.keys():
                if var_name == 'EMISSION_SPECTRA':
                    self._variables[var_name].value = load_obj.variables[var_name]
                    continue
                if var_name in ['DATETIME', 'STARTING_TIME', 'STOPPING_TIME',
                                'FILE_VERSION', 'DATA_PRODUCT_VERSION', 'SOFTWARE_VERSION_NUMBER',
                                'CALIBRATION_PERIOD_VERSION']:
                    value = np.array([load_obj.variables[var_name]])[np.newaxis, :]
                else:
                    value = np.empty((1, ), dtype=object)
                    value[0] = load_obj.variables[var_name]
                    # value = np.array([[load_obj.variables[var_name]]], dtype=object)
                self._variables[var_name].join(value)

            # self.orbit_id = load_obj.metadata['ORBIT_ID']
            # self.select_beams(field_aligned=True)
        if self.time_clip:
            self.time_filter_by_range()

    def get_time_ind(self, ut, time_res=20*60, var_datetime_name='DATETIME', edge_cutoff=False, **kwargs):
        ind = super().get_time_ind(ut, time_res=time_res, var_datetime_name=var_datetime_name, edge_cutoff=edge_cutoff, **kwargs)
        return ind
    
    def regriddata(self, disk_data=None, disk_geo_lat=None, disk_geo_lon=None, *, across_res=20., interp_method='linear'):
        from scipy.interpolate import interp1d, griddata
        
        data_pts = disk_data.flatten()
        ind_valid = np.where(np.isfinite(data_pts))[0]
        xd = range(disk_data.shape[0])
        yd = range(disk_data.shape[1])

        # create the new grids
        ps_across = self['ACROSS_PIXEL_SIZE'].value[0]
        xx = range(disk_data.shape[0])
        yy = []
        for ind in range(disk_data.shape[1]-1):
            ps1 = ps_across[ind]
            ps2 = ps_across[ind+1]
            n_insert = int(np.round((ps1 + ps2) / 2 / across_res) - 1)
            yy.extend(np.linspace(ind, ind+1-0.001, n_insert + 2)[0: -1])
        yy.extend([ind+1])
        # grid_x, grid_y = np.meshgrid(xx, yy, indexing='ij')
        
        factor = np.pi / 180.
        sin_glat = np.sin(disk_geo_lat * factor)
        itpf_sin = interp1d(yd, sin_glat, kind='cubic', bounds_error=False, fill_value='extrapolate')
        sin_glat_i = itpf_sin(yy)
        sin_glat_i[sin_glat_i>1.] = 1.
        sin_glat_i[sin_glat_i<-1.] = -1. 
        
        cos_glat = np.cos(disk_geo_lat * factor)
        itpf_cos = interp1d(yd, cos_glat, kind='cubic', bounds_error=False, fill_value='extrapolate')
        cos_glat_i = itpf_cos(yy)
        cos_glat_i[cos_glat_i>1.] = 1.
        cos_glat_i[cos_glat_i<-1.] = -1.  
        
        rad = np.sign(sin_glat_i) * (np.pi / 2 - np.arcsin(cos_glat_i))
        grid_lat = rad / factor
        
        sin_glon = np.sin(disk_geo_lon * factor)
        itpf_sin = interp1d(yd, sin_glon, kind='cubic', bounds_error=False, fill_value='extrapolate')
        sin_glon_i = itpf_sin(yy)
        sin_glon_i[sin_glon_i>1.] = 1.
        sin_glon_i[sin_glon_i<-1.] = -1. 
        
        cos_glon = np.cos(disk_geo_lon * factor)
        itpf_cos = interp1d(yd, cos_glon, kind='cubic', bounds_error=False, fill_value='extrapolate')
        cos_glon_i = itpf_cos(yy)
        cos_glon_i[cos_glon_i>1.] = 1.
        cos_glon_i[cos_glon_i<-1.] = -1.  
        
        rad = np.sign(sin_glon_i) * (np.pi / 2 - np.arcsin(cos_glon_i))
        grid_lon = rad / factor 
        
        itpf_data = interp1d(yd, disk_data, kind='linear', bounds_error=False, fill_value='extrapolate')
        grid_data = itpf_data(yy)
        return grid_lat, grid_lon, grid_data
        

    def search_data_files(self, **kwargs):
        dt_fr = self.dt_fr
        if self.dt_to.hour > 22:
            dt_to = self.dt_to + datetime.timedelta(days=1)
        else:
            dt_to = self.dt_to
        diff_days = dttool.get_diff_days(dt_fr, dt_to)
        dt0 = dttool.get_start_of_the_day(dt_fr)
        for i in range(diff_days + 1):
            thisday = dt0 + datetime.timedelta(days=i)
            initial_file_dir = kwargs.pop('initial_file_dir', None)
            
            if initial_file_dir is None:
                initial_file_dir = self.data_root_dir / self.sat_id.upper() / str(thisday.year) /thisday.strftime("%Y%m%d")

            file_patterns = [
                'dmsp' + self.sat_id.lower(),
                'sdr-disk',
                thisday.strftime("%Y%j") + 'T',
            ]
            
            if self.orbit_id is not None:
                file_patterns.extend(['REV', self.orbit_id])

            # remove empty str
            file_patterns = [pattern for pattern in file_patterns if str(pattern)]

            search_pattern = '*' + '*'.join(file_patterns) + '*'

            if self.orbit_id is not None:
                multiple_files = False
            else:
                fp_log = initial_file_dir / (self.product.upper() + '.full.log')
                if not fp_log.is_file():
                    self.download_data(dt_fr=thisday, dt_to=thisday)
                multiple_files = True
            done = super().search_data_files(
                initial_file_dir=initial_file_dir,
                search_pattern=search_pattern,
                allow_multiple_files=multiple_files,
            )
            if done and self.orbit_id is not None:
                return True

            # Validate file paths

            if (not done and self.allow_download) or (self.force_download):
                done = self.download_data(dt_fr=thisday, dt_to=thisday)
                if done:
                    done = super().search_data_files(
                        initial_file_dir=initial_file_dir,
                        search_pattern=search_pattern,
                        allow_multiple_files=multiple_files
                    )

        return done

    def download_data(self, dt_fr=None, dt_to=None):
        if dt_fr is None:
            dt_fr = self.dt_fr
        if dt_to is None:
            dt_to = self.dt_to
        download_obj = self.downloader(
            dt_fr, dt_to,
            orbit_id=self.orbit_id, sat_id=self.sat_id,
            root_dir_local=self.data_root_dir,
            force_download=self.force_download,
            dry_run=self.download_dry_run,
        )
        return any(download_obj.done)

    @property
    def database(self):
        return self._database

    @database.setter
    def database(self, value):
        if isinstance(value, str):
            self._database = DatabaseModel(value)
        elif issubclass(value.__class__, DatabaseModel):
            self._database = value
        else:
            raise TypeError

    @property
    def product(self):
        return self._product

    @product.setter
    def product(self, value):
        if isinstance(value, str):
            self._product = ProductModel(value)
        elif issubclass(value.__class__, ProductModel):
            self._product = value
        else:
            raise TypeError

    @property
    def facility(self):
        return self._facility

    @facility.setter
    def facility(self, value):
        if isinstance(value, str):
            self._facility = FacilityModel(value)
        elif issubclass(value.__class__, FacilityModel):
            self._facility = value
        else:
            raise TypeError

    @property
    def instrument(self):
        return self._instrument

    @instrument.setter
    def instrument(self, value):
        if isinstance(value, str):
            self._instrument = InstrumentModel(value)
        elif issubclass(value.__class__, InstrumentModel):
            self._instrument = value
        else:
            raise TypeError



# Licensed under the BSD 3-Clause License
# Copyright (C) 2021 GeospaceLab (geospacelab)
# Author: Lei Cai, Space Physics and Astronomy, University of Oulu

__author__ = "Lei Cai"
__copyright__ = "Copyright 2021, GeospaceLab"
__license__ = "BSD-3-Clause License"
__email__ = "lei.cai@oulu.fi"
__docformat__ = "reStructureText"


import time
import datetime
import re
import pathlib
import madrigalWeb.madrigalWeb as madrigalweb
import numpy as np
import copy


from geospacelab.config import prf
import geospacelab.datahub.sources.madrigal as madrigal
import geospacelab.toolbox.utilities.pylogging as mylog
import geospacelab.toolbox.utilities.pydatetime as dttool


DEFAULT_MADRIGAL_URL = "http://cedar.openmadrigal.org/"


class Downloader(object):

    def __init__(
            self, dt_fr: datetime.datetime, dt_to: datetime,
            icodes: list,
            include_exp_name_patterns: list=None,
            exclude_exp_name_patterns: list=None,
            include_exp_ids: list = None,
            exclude_exp_ids: list= None,
            include_file_name_patterns: list = None,
            exclude_file_name_patterns: list = None,
            include_file_type_patterns=None,
            exclude_file_type_patterns=None,
            data_file_root_dir: str = None,
            direct_download = False,
            force_download = False,
            dry_run: bool=False,
            madrigal_url: str = DEFAULT_MADRIGAL_URL,
            user_fullname: str=madrigal.default_user_fullname,
            user_email: str=madrigal.default_user_email,
            user_affiliation: str=madrigal.default_user_affiliation):
        """

        :param dt_fr: Starting time.
        :type dt_fr: datetime.datetime
        :param dt_to: Stopping time.
        :type dt_to: datetime.datetime
        :param data_file_root_dir: The root directory for the data.
        :type data_file_root_dir: str  or pathlib.Path
        :param user_fullname:
        :param user_email:
        :param user_affiliation:
        """

        self.user_fullname = user_fullname
        self.user_email = user_email
        self.user_affiliation = user_affiliation

        self.dt_fr = dt_fr
        self.dt_to = dt_to

        self.icodes = icodes
        self.madrigal_url = madrigal_url

        self.dry_run = dry_run
        self.include_exp_ids = include_exp_ids if include_exp_ids is not None else []
        self.exclude_exp_ids = exclude_exp_ids if exclude_exp_ids is not None else []
        self.include_exp_name_patterns = include_exp_name_patterns \
            if include_exp_name_patterns is not None else []
        self.exclude_exp_name_patterns = exclude_exp_name_patterns \
            if exclude_exp_name_patterns is not None else []
        self.include_file_name_patterns = include_file_name_patterns if include_file_name_patterns is not None else []
        self.exclude_file_name_patterns = exclude_file_name_patterns if exclude_file_name_patterns is not None else []
        self.include_file_type_patterns = include_file_type_patterns if include_file_type_patterns is not None else []
        self.exclude_file_type_patterns = exclude_file_type_patterns if exclude_file_type_patterns is not None else []

        self.force_download = force_download
        
        self.data_file_root_dir = pathlib.Path(data_file_root_dir) if isinstance(data_file_root_dir, str) else data_file_root_dir
        if self.data_file_root_dir is None:
            self.data_file_root_dir = pathlib.Path().absolute()
        
        self.exp_list = []
        self.exp_list_error = []
        self.database = None
        
        self.data_file_paths = []

        self.done = False
        
        if dry_run:
            self.show_all_experiment_and_files()
            return
        if direct_download:
            self.download()

    def show_all_experiment_and_files(self):
        exps, database = self.get_exp_list(
            exp_ids=[], exp_name_patterns=[], display=False)
        exps = self.get_online_file_list(exp_list=exps, display=True)
        return

    def download(self, database=None, file_path_remote=None, file_path_local=None, file_format='hdf5'):
        database = self.database if database is None else database
        self.data_file_paths.append(file_path_local)
        
        if file_path_local.is_file():
            mylog.simpleinfo.info("The file {} has been downloaded.".format(file_path_local.name))
            if not self.force_download:
                self.done = True
                return
        files_error = []
        mylog.simpleinfo.info("Downloading {} ...".format(file_path_remote))
        try:
            database.downloadFile(
                file_path_remote, file_path_local,
                self.user_fullname, self.user_email, self.user_affiliation,
                file_format
            )

            mylog.simpleinfo.info("--> Saved as {}.".format(file_path_local))
            self.done = True
        except Exception as e:
            try:
                time.sleep(3)
                database.downloadFile(
                    file_path_remote, file_path_local,
                    self.user_fullname, self.user_email, self.user_affiliation,
                    file_format
                )

                mylog.simpleinfo.info("--> Saved as {}.".format(file_path_local))
                self.done = True 
            except:
                print(e)
                mylog.StreamLogger.warning(f"Failed to download the file: {file_path_remote}")
        return
    
    @staticmethod
    def get_online_file_list(
            exp_list=None,
            include_file_name_patterns=None,
            exclude_file_name_patterns=None,
            include_file_type_patterns=None,
            exclude_file_type_patterns=None,
            database=None, display=False):

        def try_to_get_experiment_files(max=3, interval=10):
            for m in range(max):
                try: 
                    files = database.getExperimentFiles(exp.id)
                    return files
                except Exception as e:
                    if m < max - 1:
                        time.sleep(interval)
                    continue
            print(e)
            mylog.StreamLogger.warning(f"Failed to get experiment files with {max} connection(s)!")
            return -1
        
        include_file_name_patterns = [] if include_file_name_patterns is None else include_file_name_patterns
        exclude_file_name_patterns = [] if exclude_file_name_patterns is None else exclude_file_name_patterns
        include_file_type_patterns = [] if include_file_type_patterns is None else include_file_type_patterns
        exclude_file_type_patterns = [] if exclude_file_type_patterns is None else exclude_file_type_patterns

        exps_new = []
        exps_error = []
        mylog.simpleinfo.info("Searching files ...")
        for exp in exp_list:
            time.sleep(1)
            mylog.simpleinfo.info(f"Checking the experiment: {exp.name} (ID: {exp.id})")
            
            files = try_to_get_experiment_files(max=3, interval=10)
            
            if files == -1:
                mylog.StreamLogger.warning(
                    "Error when querying files: {} (ID: {}) in {}".format(exp.name, exp.id, exp.url)
                )
                exps_error.append(exp)
                continue

            if list(include_file_name_patterns):
                files_new = []
                for file in files:
                    matching = 0
                    for fnp in include_file_name_patterns:
                        if isinstance(fnp, str):
                            fnp = [fnp]
                        if isinstance(fnp, list):
                            fnp = r'.*' + '.*'.join(fnp) + '.*'
                        else:
                            raise AttributeError
                        rc = re.compile(fnp)
                        file_name = pathlib.Path(file.name).name
                        rm = rc.match(file_name.lower())
                        if rm is not None:
                            matching = 1
                    if matching == 1:
                        files_new.append(file)
                if not list(files_new):
                    mylog.StreamLogger.warning(
                        f"No files matching the file name patterns! Experiment: {exp.name} (ID: {exp.id}).")
                    continue
                files = np.array(files_new)

            if list(exclude_file_name_patterns):
                files_new = []
                for file in files:
                    matching = 1
                    for fnp in exclude_file_name_patterns:
                        if isinstance(fnp, str):
                            fnp = [fnp]
                        if isinstance(fnp, list):
                            fnp = r'.*' + '.*'.join(fnp) + '.*'
                        else:
                            raise AttributeError
                        rc = re.compile(fnp)
                        file_name = pathlib.Path(file.name).name
                        rm = rc.match(file_name.lower())
                        if rm is not None:
                            matching = 0
                    if matching == 1:
                        files_new.append(file)
                if not list(files_new):
                    mylog.StreamLogger.warning(
                        f"All files are excluded with the file name patterns! Experiment: {exp.name} (ID: {exp.id}).")
                    continue
                files = np.array(files_new)

            if list(include_file_type_patterns):
                files_new = []
                for file in files:
                    matching = 0
                    for fnp in include_file_type_patterns:
                        if isinstance(fnp, str):
                            fnp = [fnp]
                        if isinstance(fnp, list):
                            fnp = r'.*' + '.*'.join(fnp) + '.*'
                        else:
                            raise AttributeError
                        rc = re.compile(fnp)
                        rm = rc.match(file.kindatdesc.lower())
                        if rm is not None:
                            matching = 1
                    if matching == 1:
                        files_new.append(file)
                if not list(files_new):
                    mylog.StreamLogger.warning(
                        f"No files matching the file type patterns! Experiment: {exp.name} (ID: {exp.id}).")
                    continue
                files = np.array(files_new)

            if list(exclude_file_type_patterns):
                files_new = []
                for file in files:
                    matching = 1
                    for fnp in exclude_file_type_patterns:
                        if isinstance(fnp, str):
                            fnp = [fnp]
                        if isinstance(fnp, list):
                            fnp = r'.*' + '.*'.join(fnp) + '.*'
                        else:
                            raise AttributeError
                        rc = re.compile(fnp)
                        rm = rc.match(file.kindatdesc.lower())
                        if rm is not None:
                            matching = 0
                    if matching == 1:
                        files_new.append(file)
                if not list(files_new):
                    mylog.StreamLogger.warning(
                        f"All files are excluded with the file type patterns! Experiment: {exp.name} (ID: {exp.id}).")
                    continue
                files = np.array(files_new)

            exp.files = files
            exps_new.append(exp)

            mylog.simpleinfo.info('Listing matched files ...')
            for file in files:
                mylog.simpleinfo.info(file.name)

        if not list(exps_new):
            mylog.StreamLogger.warning(f"No experiments have the matched files!")
            return
        exps = np.array(exps_new)


        if display:
            mylog.simpleinfo.info("Listing matched experiments and files ...")
            exp_info = Downloader.get_exp_info(exps, include_file_info=True)
            mylog.simpleinfo.info("{:>10s}\t{:<24s}\t{:<24s}\t{:<16s}\t{:<15s}\t{:<40.40s}\t{:<50.50s}\t{:<80.80s}".format(
                'EXP NUM', 'START TIME', 'END TIME', 'DURATION (hour)', 'EXP ID', 'EXP Name', 'File Name', 'File Type'
            )
            )
            for ind, (exp_id, name, dt_fr, dt_to, duration) in enumerate(
                    zip(
                        exp_info['ID'], exp_info['NAME'], exp_info['DATETIME_0'], exp_info['DATETIME_1'],
                        exp_info['DURATION']
                    )
            ):
                
                for file in exp_info['FILES'][ind]:
                    line_str = "{:>10d}\t{:<24s}\t{:<24s}\t{:<16.1f}\t{:<15d}\t{:<40.40s}\t{:<50.50s}\t{:<80.80s}".format(
                        ind + 1,
                        dt_fr.strftime("%Y-%m-%d %H:%M:%S"),
                        dt_to.strftime("%Y-%m-%d %H:%M:%S"),
                        duration,
                        exp_id,
                        name,
                        file['NAME'],
                        file['DESC']
                    )
                    mylog.simpleinfo.info(line_str)
            mylog.simpleinfo.info("")
        return exps, exps_error

    @staticmethod
    def get_exp_list(
            dt_fr=None, dt_to=None,
            include_exp_name_patterns=None,
            exclude_exp_name_patterns=None,
            include_exp_ids=None,
            exclude_exp_ids=None,
            icodes=None, madrigal_url=None, display=True, level=0,
    ): 
        
        def try_to_get_database(max=3, interval=30):
            for m in range(max):
                try: 
                    database = madrigalweb.MadrigalData(madrigal_url)
                    return database
                except Exception as e:
                    if m < max - 1:
                        time.sleep(interval)
                    continue
            
            print(e)
            raise ImportError(f"Failed to connect the Madrigal database with {max} connection(s)!")
        
        def try_to_get_experiments(max=3, interval=30):
            for m in range(max):
                try: 
                    exps_o = database.getExperiments(
                        icode,
                        dt_fr.year, dt_fr.month, dt_fr.day, dt_fr.hour, dt_fr.minute, dt_fr.second,
                        dt_to.year, dt_to.month, dt_to.day, dt_to.hour, dt_to.minute, dt_to.second,
                        local=0
                    )
                    return exps_o
                except Exception as e:
                    if m < max - 1:
                        time.sleep(interval)
                    continue
            
            print(e)
            raise ImportError(f"Failed to get the experiments from the database with {max} connection(s)!") 
        
        include_exp_name_patterns = [] if include_exp_name_patterns is None else include_exp_name_patterns
        include_exp_ids = [] if include_exp_ids is None else include_exp_ids
        icodes = [] if icodes is None else icodes
        madrigal_url = DEFAULT_MADRIGAL_URL if madrigal_url is None else madrigal_url
        exclude_exp_ids = [] if exclude_exp_ids is None else exclude_exp_ids
        
        exps = []
        mylog.simpleinfo.info(f"Contacting the Madrigal database (URL: {madrigal_url}) ...")
        database = try_to_get_database(max=3, interval=30) 
            
        for icode in icodes:
            mylog.simpleinfo.info("Searching experiments ...")
            exps_o = try_to_get_experiments(max=3, interval=30)
            exps.extend(exps_o)
        exps = np.array(exps, dtype=object)

        exps_new = []
        another_madrigal_url = ''
        
        for exp in exps:
            if exp.id == -1 and level==0:
                if another_madrigal_url != exp.madrigalUrl:
                    mylog.StreamLogger.warning(
                        f'Another Madrigal site detected: {exp.madrigalUrl}!'
                    )
                    another_madrigal_url = exp.madrigalUrl
                continue
            if exp.id in exclude_exp_ids:
                mylog.StreamLogger.warning(
                    f'The following experiment is excluded: {exp.name} (ID: {exp.id})!'
                )
                continue
            exps_new.append(exp)
  
        exps = np.array(exps_new, dtype=object)

        if not list(exps) and not str(another_madrigal_url):
            raise ValueError('Cannot find available experiments from the current database! Check the input values!')
        elif str(another_madrigal_url) and level==0:
            mylog.StreamLogger.warning(
                'Some data are located in another Madrigal site: {}.'.format(another_madrigal_url)
            )
            
            exps, database = Downloader.get_exp_list(
                    dt_fr=dt_fr, dt_to=dt_to,
                    include_exp_name_patterns=include_exp_name_patterns,
                    exclude_exp_name_patterns=exclude_exp_name_patterns,
                    include_exp_ids=include_exp_ids,
                    exclude_exp_ids=exclude_exp_ids,
                    icodes=icodes, madrigal_url=another_madrigal_url, display=display, level=level+1
            ) 
        else:
            pass
        
        dts_fr = np.array([
                datetime.datetime(
                    exp.startyear, exp.startmonth, exp.startday, exp.starthour, exp.startmin, exp.startsec)
                for exp in exps
            ])
        dts_to = np.array([
            datetime.datetime(
                exp.endyear, exp.endmonth, exp.endday, exp.endhour, exp.endmin, exp.endsec)
            for exp in exps
        ])

        eids = np.array([exp.id for exp in exps])

        if list(include_exp_ids):
            exps_new = []
            for exp_id in include_exp_ids:
                ind = np.where(eids == exp_id)[0]

                if list(ind):
                    exps_new.append(exps[ind[0]])
                else:
                    mylog.StreamLogger.warning(
                        f'The requested experiment (ID: {exp_id}) cannot be found!'
                    )
            exps = np.array(exps_new, dtype=object)
            # inds_o = np.array(eids).argsort()
            # inds = inds_o[np.searchsorted(eids[inds_o], include_exp_ids)]
            # if not list(inds):
            #     mylog.StreamLogger.error("Cannot find available experiments for the input experiment IDs!")
            #     raise AttributeError
            # exps = exps[inds]

        else:
            ind_dt_no = np.where(
                (dt_fr > dts_to) | (dt_to < dts_fr))[0]
            exps = exps[[i not in ind_dt_no for i in range(eids.size)]]
            if not list(exps):
                mylog.StreamLogger.error("No experiments matching the time range!")
                raise AttributeError

            if list(include_exp_name_patterns):
                exps_new = []
                for exp in exps:
                    matching = 0
                    for enp in include_exp_name_patterns:
                        if isinstance(enp, str):
                            enp = [enp]
                        if isinstance(enp, list):
                            enp = r'.*' + '.*'.join(enp) + '.*'
                        else:
                            raise AttributeError
                        rc = re.compile(enp)
                        rm = rc.match(exp.name.lower())
                        if rm is not None:
                            matching = 1
                    if matching == 1:
                        exps_new.append(exp)
                if not list(exps_new):
                    mylog.StreamLogger.warning(
                        f"No experiments matching the exp name patterns!")
                    raise AttributeError
                exps = np.array(exps_new, dtype=object)

            if list(exclude_exp_name_patterns):
                exps_new = []
                for exp in exps:
                    matching = 1
                    for enp in exclude_exp_name_patterns:
                        if isinstance(enp, str):
                            enp = [enp]
                        if isinstance(enp, list):
                            enp = r'.*' + '.*'.join(enp) + '.*'
                        else:
                            raise AttributeError
                        rc = re.compile(enp)
                        rm = rc.match(exp.name.lower())
                        if rm is not None:
                            matching = 0
                    if matching == 1:
                        exps_new.append(exp)
                if not list(exps_new):
                    mylog.StreamLogger.warning(
                        f"All experiments are excluded with the exp name patterns!")
                    raise AttributeError
                exps = np.array(exps_new, dtype=object)

            if not list(exps):
                mylog.StreamLogger.error("Cannot find available experiments for the input experiment name patterns!")
                raise AttributeError

        if display:
            mylog.simpleinfo.info("Listing matched experiments ...")
            exp_info = Downloader.get_exp_info(exps)
            mylog.simpleinfo.info(
                "{:>10s}\t{:<24s}\t{:<24s}\t{:<16s}\t{:<15s}\t{:<40.40s}\t{:<s}".format(
                'EXP NUM', 'START TIME', 'END TIME', 'DURATION (hour)', 'EXP ID', 'EXP Name', 'EXP_URL'
                )
            )
            for ind, (exp_id, name, dt_fr, dt_to, duration, url) in enumerate(
                    zip(
                        exp_info['ID'], exp_info['NAME'], exp_info['DATETIME_0'], exp_info['DATETIME_1'],
                        exp_info['DURATION'], exp_info['URL']
                    )
            ):
                line_str = "{:>10d}\t{:<24s}\t{:<24s}\t{:<16.1f}\t{:<15d}\t{:<40.40s}\t{:<s}".format(
                    ind+1,
                    dt_fr.strftime("%Y-%m-%d %H:%M:%S"),
                    dt_to.strftime("%Y-%m-%d %H:%M:%S"),
                    duration,
                    exp_id,
                    name,
                    url
                )

                mylog.simpleinfo.info(line_str)
            mylog.simpleinfo.info("")

        return exps, database

    @staticmethod
    def get_exp_info(exps, include_file_info=False):

        exp_info = {
                'ID': [],
                'URL': [],
                'NAME': [],
                'DURATION': [],
                'DATETIME_0': [],
                'DATETIME_1': [],
            }
        if include_file_info:
            exp_info['FILES'] = []
        file_info = {
            'NAME': '',
            'DESC': '',
            'STATUS': '',
        }

        for exp in exps:
            dt_fr = datetime.datetime(
                exp.startyear, exp.startmonth, exp.startday, exp.starthour, exp.startmin, exp.startsec
            )
            dt_to = datetime.datetime(
                exp.endyear, exp.endmonth, exp.endday, exp.endhour, exp.endmin, exp.endsec
            )
            duration = (dt_to - dt_fr).total_seconds() / 3600

            exp_info['ID'].append(exp.id)
            exp_info['URL'].append(exp.url)
            exp_info['NAME'].append(exp.name)
            exp_info['DATETIME_0'].append(dt_fr)
            exp_info['DATETIME_1'].append(dt_to)
            exp_info['DURATION'].append(duration)

            if include_file_info:
                exp_info['FILES'].append([])
                for file in exp.files:
                    file_path_remote = pathlib.Path(file.name)
                    file_name = file_path_remote.name
                    exp_info['FILES'][-1].append(copy.deepcopy(file_info))
                    exp_info['FILES'][-1][-1]['NAME'] = file_name
                    exp_info['FILES'][-1][-1]['DESC'] = file.kindatdesc
                    exp_info['FILES'][-1][-1]['STATUS'] = file.status

        return exp_info


# Licensed under the BSD 3-Clause License
# Copyright (C) 2021 GeospaceLab (geospacelab)
# Author: Lei Cai, Space Physics and Astronomy, University of Oulu

__author__ = "Lei Cai"
__copyright__ = "Copyright 2021, GeospaceLab"
__license__ = "BSD-3-Clause License"
__email__ = "lei.cai@oulu.fi"
__docformat__ = "reStructureText"


import h5py
import numpy as np
import cftime
import re
import datetime

import geospacelab.toolbox.utilities.pylogging as mylog
import geospacelab.toolbox.utilities.pydatetime as dttool


var_name_dict = {
    'AZ': 'azm',
    'EL': 'elm',
    'PULSE_LENGTH': 'pl',
    'P_Tx': 'power',
    'n_e': 'ne',
    'n_e_err': 'dne',
    'T_i': 'ti',
    'T_i_err': 'dti',
    'T_e': 'te',
    'T_e_err': 'dte',
    'v_i_los': 'vo',
    'v_i_los_err': 'dvo',
    'comp_O_p': 'po+',
    'comp_O_p_err': 'dpo+',
    'CGM_LAT': 'cgm_lat',
    'CGM_LON': 'cgm_long',
    'HEIGHT': 'gdalt',
    'BEAM_ID': 'beamid',
    'CHISQ': 'chisq',
}


class Loader:
    """
    :param file_path: the file's full path
    :type file_path: pathlib.Path object
    :param file_type: the specific file type for the file being loaded. Options: ['TEC-MAT'], 'TEC-LOS', 'TEC-sites')
    :type file_type: str
    :param load_data: True, load without calling the method "load_data" separately.
    :type load_data: bool
    """
    def __init__(self, file_path, beam_id=None, beam_az=None, beam_el=None, direct_load=True):
        self.file_path = file_path
        self.beam_id = beam_id
        self.beam_az = beam_az
        self.beam_el = beam_el
        self.beams = None
        self.variables = {}
        self.metadata = {}

        self.done = False
        if direct_load:
            self.load()
    
    def list_beams(self, display=True):
        with h5py.File(self.file_path, 'r') as fh5:
            data_fh5 = fh5['Data']
            array_layouts = list(data_fh5['Array Layout'].keys())
            
            beam_ids = []
            beam_azs = []
            beam_els = []
            for i, array_layout_str in enumerate(array_layouts):
                
                res = re.search(r'beamid=([\d.]+)', array_layout_str)
                beam_id = int(float(res.groups()[0]))
                beam_az = np.nanmedian(data_fh5['Array Layout'][array_layout_str]['1D Parameters']['azm'])
                beam_el = np.nanmedian(data_fh5['Array Layout'][array_layout_str]['1D Parameters']['elm'])

                beam_ids.append(beam_id)
                beam_azs.append(beam_az)
                beam_els.append(beam_el)
        beam_azs = [beam_az % 360 for beam_az in beam_azs]
        beam_data = {
            'IDs': beam_ids,
            'AZs': beam_azs,
            'ELs': beam_els,
        }
        self.beams = beam_data
        
        if display:
            mylog.simpleinfo.info("Listing the beams ...")
            mylog.simpleinfo.info("{:>10s}{:>12s}{:>12.2s}{:>12.2s}".format('Num', 'ID', 'AZ', 'EL'))
            for i, beam_id in enumerate(beam_ids):
                mylog.simpleinfo.info(
                    "{:>10d}{:>12d}{:>12.2f}{:>12.2f}".format(
                        i + 1,
                        beam_id,
                        beam_azs[i],
                        beam_els[i]
                    )
                )
        return 
                
    def select_beam(self, az=None, el=None, error_az=0.2, error_el=0.2):
        
        beam_azs = np.array(self.beams['AZs'])
        beam_els = np.array(self.beams['ELs'])
        beam_ids = np.array(self.beams['IDs'])
        
        ind_1 = np.where(((np.abs(beam_azs - az) <= error_az) & (np.abs(beam_els - el) <= error_el)))[0]
        ind_2 = np.where(((np.abs(beam_azs - 360. - az) <= error_az) & (np.abs(beam_els - el) <= error_el)))[0] 
        ind_1 = np.append(ind_1, ind_2) 
        
        if not list(ind_1):
            mylog.StreamLogger.error(f"Cannot find the beam with AZ={az} and EL={el}!")
            raise AttributeError
        
        if len(ind_1) > 2:
            mylog.StreamLogger.error(f"Multiple beams found to fulfill AZ={az} and EL={el}!")
            raise AttributeError
        
        self.beam_id = beam_ids[ind_1[0]]
        self.beam_az = beam_azs[ind_1[0]]
        self.beam_el = beam_els[ind_1[0]]
        mylog.simpleinfo.info(f'Select the beam: ID={self.beam_id}, AZ={self.beam_az}, EL={self.beam_el}')
        return self.beam_id
    
    def get_metadata(self):
        with h5py.File(self.file_path, 'r') as fh5:
            metadata_fh5 = fh5['Metadata'] 
                
            exp_data = metadata_fh5['Experiment Parameters'][::]
            metadata = {}
            for (key, value) in exp_data:
                metadata[key.decode('UTF-8').strip()] = value.decode('UTF-8').strip()
             
            metadata['NOTES'] = [item[0].decode('UTF-8') for item in metadata_fh5['Experiment Notes'][::]]
        
        return metadata

    def load_from_table_layout(self):
        variables = {}
        metadata = {}



    def load(self):
        variables = {}
        metadata = {}
        
        if self.beam_id is None:
            self.list_beams(display=True)
            
            self.select_beam(az=self.beam_az, el=self.beam_el)
        
        with h5py.File(self.file_path, 'r') as fh5:
            data_fh5 = fh5['Data']
            array_layouts = list(data_fh5['Array Layout'].keys())

            matching = 0
            for array_layout_str in array_layouts:
                if str(self.beam_id) in array_layout_str:
                    matching = 1
                    break
                
            if matching == 0:
                raise AttributeError

            vars_fh5 = {}
            fh5_vars_1d = data_fh5['Array Layout'][array_layout_str]['1D Parameters']
            for var_name in fh5_vars_1d.keys():
                if var_name == 'Data Parameters':
                    continue
                vars_fh5[var_name] = np.array(fh5_vars_1d[var_name])[:, np.newaxis]
            fh5_vars_2d = data_fh5['Array Layout'][array_layout_str]['2D Parameters']
            for var_name in fh5_vars_2d.keys():
                if var_name == 'Data Parameters':
                    continue
                if var_name == 'nel':
                    vars_fh5['ne'] = 10**np.array(fh5_vars_2d[var_name]).T
                elif var_name == 'dnel':
                    vars_fh5['dne'] = 10 ** np.array(fh5_vars_2d[var_name]).T
                else:
                    vars_fh5[var_name] = np.array(fh5_vars_2d[var_name]).T
            vars_fh5['range'] = np.array(data_fh5['Array Layout'][array_layout_str]['range'])[np.newaxis, :]
            if np.median(vars_fh5['range'].flatten()) > 1e5:
                mylog.StreamLogger.warning(f"The variable range is detected in [m]. It is converted into [km].")
                vars_fh5['range'] = vars_fh5['range'] * 1e-3
            vars_fh5['timestamps'] = np.array(data_fh5['Array Layout'][array_layout_str]['timestamps'])[:, np.newaxis]
            for var_name, var_name_fh5 in var_name_dict.items():
                if var_name_fh5 not in vars_fh5.keys():
                    mylog.StreamLogger.warning(f"The requested variable {var_name_fh5} does not exist in the data file!")
                    variables[var_name] = None
                    continue
                variables[var_name] = vars_fh5[var_name_fh5]

            variables['comp_mix'] = 1. - variables['comp_O_p']
            variables['comp_mix_err'] = variables['comp_O_p_err']

            variables['RANGE'] = np.tile(vars_fh5['range'], [variables['n_e'].shape[0], 1])
            variables['DATETIME'] = dttool.convert_unix_time_to_datetime_cftime(vars_fh5['timestamps'])
                
            self.variables = variables
            
            metadata = self.get_metadata()
            
            if 'alternating' in metadata['kind of data file'].lower():
                pulse_code = 'Alternating Code'
            elif 'long pulse' in metadata['kind of data file'].lower():
                pulse_code = 'Long Pulse'
            else:
                pulse_code = ''
            
            pulse_length = np.nanmedian(variables['PULSE_LENGTH'].flatten())
            
            metadata['PULSE_CODE'] = pulse_code
            metadata['PULSE_LENGTH'] = pulse_length

            self.beam_az = np.nanmedian(variables['AZ'])
            self.beam_el = np.nanmedian(variables['EL'])
            
            self.metadata = metadata



from geospacelab.config import pref
import geospacelab.datahub.sources.madrigal.utilities as utilities

from geospacelab.datahub import DatabaseModel


class GFZDatabase(DatabaseModel):
    def __new__(cls, str_in, **kwargs):
        obj = super().__new__(cls, str_in, **kwargs)
        return obj


gfz_database = GFZDatabase('WDC')
gfz_database.url = 'https://www.gfz-potsdam.de/en/section/geomagnetism/data-products-services/'
gfz_database.category = 'online database'
gfz_database.Notes = ''



# Licensed under the BSD 3-Clause License
# Copyright (C) 2021 GeospaceLab (geospacelab)
# Author: Lei Cai, Space Physics and Astronomy, University of Oulu

__author__ = "Lei Cai"
__copyright__ = "Copyright 2021, GeospaceLab"
__license__ = "BSD-3-Clause License"
__email__ = "lei.cai@oulu.fi"
__docformat__ = "reStructureText"

import numpy as np
import copy
from geospacelab.datahub import VariableModel as Var
import geospacelab.visualization.mpl.colormaps as cm

database = 'Madrigal'

timestamps = {
    'DATETIME': 'SC_DATETIME',
}


coords = {
    'GEO': ['SC_GEO_LAT', 'SC_GEO_LON', 'SC_GEO_ALT', 'SC_GEO_ST'],
    'AACGM': ['SC_AACGM_LAT', 'SC_AACGM_LON', 'SC_AACGM_R', 'SC_AACGM_MLT'],
    'APEX': ['SC_APEX_LAT', 'SC_APEX_LON', 'SC_APEX_ALT', 'SC_APEX_MLT']
}

depend_0 = {
    'UT': 'SC_DATETIME',
    'GEO_LAT': 'SC_GEO_LAT', 'GEO_LON': 'SC_GEO_LON',
    'AACGM_LAT': 'SC_AACGM_LAT', 'AACGM_LON': 'SC_AACGM_LON', 'AACGM_MLT': 'SC_AACGM_MLT'
}


default_colormap = cm.cmap_jet_modified()

default_axis_dict_1d = {
    1: {
        'data': '@v.value',
        'label': '@v.group',
        'unit': '@v.unit_label',
        'label_pos': [-0.1, 0.5],
    },
    2: {
        'label': '@v.label'
    },
}

default_axis_dict_2d = {
    1:     {
        'data':     '@d.HEIGHT.value',
        'lim':      [90, 350],
        'label':    'h',
        'unit':     'km',
    },
    2:  {
        'data':     '@v.value',
        'label':    '@v.label',
        'unit':     '@v.unit_label',
    }
}

default_plot_config = {
    'line':         {
        'linestyle':        '-',
        'linewidth':        1.5,
        'marker':           '.',
        'markersize':       2,
    },
    'pcolormesh':   {
        'cmap':            default_colormap,
    }
}

configured_variables = {}
visual = 'on'

####################################################################################################################
var_name = 'B_D'
var = Var(ndim=1, variable_type='scalar', visual=visual)
# set variable attrs
var_config = {
    'name': var_name,
    'fullname': 'Downward magnetic field',
    'label': r'$B_D$',
    'group': r'$B$',
    'unit': 'nT',
    'unit_label': None,
    'error': None,
    'depends': {0: depend_0},
}
var.config(**var_config)
# set plot attrs
var.visual.plot_config.config(**default_plot_config)
var.visual.plot_config.style = '1noE'
# set axis attrs
var.visual.axis[1].config(**default_axis_dict_1d[1])
var.visual.axis[1].lim = [-np.inf, np.inf]
var.visual.axis[1].data_scale = 1e9

var.visual.axis[2].config(**default_axis_dict_1d[2])

configured_variables[var_name] = var

####################################################################################################################
var_name = 'B_P'
var = Var(ndim=1, variable_type='scalar', visual=visual)
# set variable attrs
var_config = {
    'name': var_name,
    'fullname': 'Cross-track magnetic field ',
    'label': r'$B_P$',
    'group': r'$B$',
    'unit': 'nT',
    'unit_label': None,
    'error': None,
    'depends': {0: depend_0},
}
var.config(**var_config)
# set plot attrs
var.visual.plot_config.config(**default_plot_config)
var.visual.plot_config.style = '1noE'
# set axis attrs
var.visual.axis[1].config(**default_axis_dict_1d[1])
var.visual.axis[1].lim = [-np.inf, np.inf]
var.visual.axis[1].data_scale = 1e9

var.visual.axis[2].config(**default_axis_dict_1d[2])

configured_variables[var_name] = var

####################################################################################################################
var_name = 'B_F'
var = Var(ndim=1, variable_type='scalar', visual=visual)
# set variable attrs
var_config = {
    'name': var_name,
    'fullname': 'Forward-track magnetic field',
    'label': r'$B_F$',
    'group': r'$B$',
    'unit': 'nT',
    'unit_label': None,
    'error': None,
    'depends': {0: depend_0},
}
var.config(**var_config)
# set plot attrs
var.visual.plot_config.config(**default_plot_config)
var.visual.plot_config.style = '1noE'
# set axis attrs
var.visual.axis[1].config(**default_axis_dict_1d[1])
var.visual.axis[1].lim = [-np.inf, np.inf]
var.visual.axis[1].data_scale = 1e9

var.visual.axis[2].config(**default_axis_dict_1d[2])

configured_variables[var_name] = var


####################################################################################################################
var_name = 'd_B_D'
var = Var(ndim=1, variable_type='scalar', visual=visual)
# set variable attrs
var_config = {
    'name': var_name,
    'fullname': 'Downward magnetic field subtracted from the background',
    'label': r'$\delta B_D$',
    'group': r'$\delta B$',
    'unit': 'nT',
    'unit_label': None,
    'error': None,
    'depends': {0: depend_0},
}
var.config(**var_config)
# set plot attrs
var.visual.plot_config.config(**default_plot_config)
var.visual.plot_config.style = '1noE'
# set axis attrs
var.visual.axis[1].config(**default_axis_dict_1d[1])
var.visual.axis[1].lim = [-np.inf, np.inf]
var.visual.axis[1].data_scale = 1e9

var.visual.axis[2].config(**default_axis_dict_1d[2])

configured_variables[var_name] = var

####################################################################################################################
var_name = 'd_B_P'
var = Var(ndim=1, variable_type='scalar', visual=visual)
# set variable attrs
var_config = {
    'name': var_name,
    'fullname': 'Cross-track magnetic field subtracted from the background',
    'label': r'$\delta B_P$',
    'group': r'$\delta B$',
    'unit': 'nT',
    'unit_label': None,
    'error': None,
    'depends': {0: depend_0},
}
var.config(**var_config)
# set plot attrs
var.visual.plot_config.config(**default_plot_config)
var.visual.plot_config.style = '1noE'
# set axis attrs
var.visual.axis[1].config(**default_axis_dict_1d[1])
var.visual.axis[1].lim = [-np.inf, np.inf]
var.visual.axis[1].data_scale = 1e9

var.visual.axis[2].config(**default_axis_dict_1d[2])

configured_variables[var_name] = var

####################################################################################################################
var_name = 'd_B_F'
var = Var(ndim=1, variable_type='scalar', visual=visual)
# set variable attrs
var_config = {
    'name': var_name,
    'fullname': 'Forward-track magnetic field subtracted from the background',
    'label': r'$\delta B_F$',
    'group': r'$\delta B$',
    'unit': 'nT',
    'unit_label': None,
    'error': None,
    'depends': {0: depend_0},
}
var.config(**var_config)
# set plot attrs
var.visual.plot_config.config(**default_plot_config)
var.visual.plot_config.style = '1noE'
# set axis attrs
var.visual.axis[1].config(**default_axis_dict_1d[1])
var.visual.axis[1].lim = [-np.inf, np.inf]
var.visual.axis[1].data_scale = 1e9

var.visual.axis[2].config(**default_axis_dict_1d[2])

configured_variables[var_name] = var




#  Install these prerequisites once before executing
#  the example code:
#    pip install -U sscws

from sscws.sscws import SscWs
ssc = SscWs()

#  Edit the following time variable to suit your needs.
time = ['2018-12-19T23:00:00.000Z', '2018-12-20T00:00:00.000Z']
result = ssc.get_locations(['dmspf18'], time)
data = result['Data'][0]
coords = data['Coordinates'][0]
print(coords['X'])
#  ...
try:
    import matplotlib as mpl
    import matplotlib.pyplot as plt
    from mpl_toolkits.mplot3d import Axes3D
    from packaging import version
    fig = plt.figure()
    if version.parse(mpl.__version__) < version.parse('3.4'):
        ax = fig.gca(projection='3d')
    else:
        ax = Axes3D(fig, auto_add_to_figure=False)
        fig.add_axes(ax)
    ax.set_xlabel('X (km)')
    ax.set_ylabel('Y (km)')
    ax.set_zlabel('Z (km)')
    title = data['Id'] + ' Orbit (' + coords['CoordinateSystem'].value.upper() + ')'
    ax.plot(coords['X'], coords['Y'], coords['Z'], label=title)
    ax.legend()
    plt.show()
except ImportError:
    print('To see the plot, do')
    print('pip install packaging matplotlib')
except Exception as e:
    print(e)



# Licensed under the BSD 3-Clause License
# Copyright (C) 2021 GeospaceLab (geospacelab)
# Author: Lei Cai, Space Physics and Astronomy, University of Oulu

__author__ = "Lei Cai"
__copyright__ = "Copyright 2021, GeospaceLab"
__license__ = "BSD-3-Clause License"
__email__ = "lei.cai@oulu.fi"
__docformat__ = "reStructureText"


import netCDF4
import numpy as np
import cftime
import datetime


class Loader:
    def __init__(self, file_path, file_type='nc', load_data=True):
        self.file_path = file_path
        self.file_type = file_type
        self.variables = {}
        self.done = False
        if load_data:
            self.load()

    def load(self):
        fnc = netCDF4.Dataset(self.file_path)
        var_names_nc = fnc.variables.keys()
        variables = {}
        for var_name_nc in var_names_nc:
            if 'smr_' in var_name_nc:
                var_name = var_name_nc.upper()
            else:
                var_name = var_name_nc
            variables[var_name] = np.array(fnc[var_name_nc]) 
            if len(variables[var_name].shape) == 1:
                variables[var_name] = variables[var_name][:, np.newaxis]

        time_units = fnc['UNIX_TIME'].units

        variables['DATETIME'] = cftime.num2date(variables['UNIX_TIME'].flatten(),
                                                units=time_units,
                                                only_use_cftime_datetimes=False,
                                                only_use_python_datetimes=True)

        variables['DATETIME'] = np.reshape(variables['DATETIME'], (fnc['UNIX_TIME'].shape[0], 1))
        self.variables = variables
        self.done = True
        fnc.close()


